local _G = getfenv(0)
local ADDON_NAME, addon = ...

-- Local versions for performance
local string = _G.string
local math = _G.math
local pairs = _G.pairs
local ipairs = _G.ipairs
local table = _G.table
local floor, ceil, min, max, abs = math.floor, math.ceil, math.min, math.max, math.abs
local tconcat, tinsert = table.concat, table.insert
local wipe = _G.wipe
local tostring = _G.tostring
local tonumber = _G.tonumber
local GetTime = _G.GetTime
local UnitDebuff = _G.UnitDebuff
local GetSpellCooldown = _G.GetSpellCooldown
local UnitHealth = _G.UnitHealth

local CompactRunes = _G.LibStub("AceAddon-3.0"):NewAddon(addon.addonName, "AceConsole-3.0", "AceEvent-3.0","AceTimer-3.0")
local L = _G.LibStub("AceLocale-3.0"):GetLocale(addon.addonName, true)
local LDB = _G.LibStub("LibDataBroker-1.1")
local LibQTip = _G.LibStub("LibQTip-1.0")
local Icon = _G.LibStub("LibDBIcon-1.0")
local LSM = _G.LibStub:GetLibrary("LibSharedMedia-3.0")

local MIN_UPDATE_TIME = addon.MIN_UPDATE_TIME

local round = addon.round
local FormatNumber = addon.FormatNumber

-- Define a simplistic class for combat statistics
local CombatStats = {}
CombatStats.__index = CombatStats

function CombatStats:new()
    local stats = {}
    _G.setmetatable(stats, CombatStats)
    stats:Reset()
    return stats
end

function CombatStats:Reset()
	self.rcProcs = 0
	self.cappedBC = 0
	self.reProcs = self.reProcs or {}
	wipe(self.reProcs)
	self.regenUses = self.regenUses or {}
	wipe(self.regenUses)
	self.autoattacks = 0
end

local RuneColors = nil
local RuneCooldownColors = nil
addon.RuneMapping = {
	[0] = "NONE",
	[1] = "BLOOD",
	[2] = "UNHOLY",
	[3] = "FROST",
	[4] = "DEATH",
}
local RuneMapping = addon.RuneMapping
addon.RuneNames = {
	["BLOOD"] = L["Blood"],
	["UNHOLY"] = L["Unholy"],
	["FROST"] = L["Frost"],
	["DEATH"] = L["Death"],
}
local RuneNames = addon.RuneNames

local HEXFMT = "|c%2x%2x%2x%2x"
local function chatColorCode(color)
	local a = 255
	local r = ceil(255 * color.r)
	local g = ceil(255 * color.g)
	local b = ceil(255 * color.b)
	return HEXFMT:format(a, r, g, b)
end

local RuneChatColors = {}
function CompactRunes:UpdateRuneColors()
	RuneColors = self.db.profile.RuneColors
	RuneCooldownColors = self.db.profile.RuneCooldownColors
	wipe(RuneChatColors)
	for k, v in pairs(self.db.profile.RuneColors) do
		RuneChatColors[k] = chatColorCode(v)
	end
end

-- Define Bar for now but the rest is at the bottom of the file.
local Bar = {}

CompactRunes.bars = {}
CompactRunes.blocks = {}
CompactRunes.cdbars = {}
CompactRunes.diseases = {
	ff = {},
	bp = {},
}
CompactRunes.playerName = UnitName("player")
CompactRunes.isDK = false
CompactRunes.TotalStats = CombatStats:new()
CompactRunes.LastFightStats = CombatStats:new()
CompactRunes.regenMethod = "None"
CompactRunes.currentSpec = ""
CompactRunes.currentProcs = {}
CompactRunes.soulReaperKnown = false
CompactRunes.outbreak = nil
addon.tierCount = { ["T15 DPS"] = 0 }
CompactRunes.currentTarget = nil
CompactRunes.targetSoulReaper = false
CompactRunes.targetHostile = false
CompactRunes.targetMaxHealth = 0
CompactRunes.targetHealth = 0
CompactRunes.targetPercent = 0
CompactRunes.srMethod = "cooldown"
CompactRunes.bloodCharges = 0
CompactRunes.bcExpires = 0
CompactRunes.altRuneBkg = nil
CompactRunes.enabled = {}
CompactRunes.ready = {}
addon.diseaseCount = 0

local LookupOrKeyMT = {__index = function (t,k) return k end}
local SpellIds = {
	["Frost Fever"] = 55095,
	["Blood Plague"] = 55078,
	["Freezing Fog"] = 59052,
	["Killing Machine"] = 51124,
	["Sudden Doom"] = 49530,
	["Blood Charge"] = 114851,
	["Blood Charge (10+)"] = 114851,  -- Used for a proc
	["Blood Charge (5-9)"] = 114851,  -- Used for a proc
	["Shadow Infusion"] = 91342,
	["Dark Transformation"] = 63560,
	["Death Coil"] = 47541,
	["Death Coil Damage"] = 47632,
	["Death Coil Heal"] = 47633,
	["Rune Strike"] = 56815,
	["Frost Strike"] = 49143,
	["Blood Tap"] = 45529,
	["Runic Empowerment"] = 81229,
	["Runic Corruption"] = 51460,
	["Dark Succor"] = 101568,
	["Crimson Scourge"] = 81141,
	["Plague Leech"] = 123693,
	["Plague Strike"] = 45462,
	["Outbreak"] = 77575,
	["Soul Reaper"] = 130735,
	["Improved Soul Reaper"] = 157342,
	["Soul Reaper (Blood)"] = 114866,
	["Soul Reaper (Frost)"] = 130735,
	["Soul Reaper (Unholy)"] = 130736,
	["Pillar of Frost"] = 51271,
	["Cinderglacier"] = 53386,
	["Auto Attack"] = 6603,
	["Bone Shield"] = 49222,
	["Blood Shield"] = 77535,
	["Summon Gargoyle"] = 49206,
	["Rune Tap"] = 48982,
	["Vampiric Blood"] = 55233,
	["Dancing Rune Weapon"] = 49028,
	["Icebound Fortitude"] = 48792,
	["Anti-Magic Shell"] = 48707,
	["Death Pact"] = 48743,
	["Army of the Dead"] = 42650,
	["Empower Rune Weapon"] = 47568,
	["Feathers of Fury"] = 138759,
	["Determination"] = 146250,
	["Surge of Strength"] = 138702,
	["Tenacious"] = 148899,
	["Outrage"] = 146245,
	["Obliterate"] = 49020,
	["Necrotic Plague"] = 155163,
	["Resolve"] = 158300,
	["Raise Dead"] = 46584,
	["Defile"] = 152280,
	["Breath of Sindragosa"] = 152279,
	["Death and Decay"] = 43265,
	["Gorefiend's Grasp"] = 108199,
	["Remorseless Winter"] = 108200,
	["Desecrated Ground"] = 108201,
	["Unholy Blight"] = 115989,
	["Lichborne"] = 49039,
	["Anti-Magic Zone"] = 51052,
	["Death's Advance"] = 96268,
	["Strangulate"] = 47476,
	["Asphyxiate"] = 108194,
	["Mind Freeze"] = 47528,
	["Festering Strike"] = 85948,
	["Obliteration"] = 187893,
	["Frozen Wake"] = 187894,
}

-- Spells which need to use the listed name and not the one from the API.
addon.SpecialSpellNames = {
	["Blood Charge (10+)"] = true,  -- Used for a proc
	["Blood Charge (5-9)"] = true,  -- Used for a proc
}

local SpellNames = {}
local SpellIcons = {}
_G.setmetatable(SpellNames, LookupOrKeyMT)
local function LoadSpellNames()
	for k, v in pairs(SpellIds) do
		local name, _, icon = _G.GetSpellInfo(v)
		SpellNames[k] = name
		SpellIcons[k] = icon
	end
end
LoadSpellNames()
addon.SpellIds = SpellIds
addon.SpellNames = SpellNames
addon.SpellIcons = SpellIcons

addon.Talents = {
	-- Level 56 talents
	["Plague Leech"] = 19166,
	["Unholy Blight"] = 19217,
	-- Level 57 talents
	["Lichborne"] = 19218,
	["Anti-Magic Zone"] = 19219,
	-- Level 58 talents
	["Death's Advance"] = 19221,
	["Asphyxiate"] = 19223,
	-- Level 60 talents
	["Blood Tap"] = 19224,
	["Runic Empowerment"] = 19225,
	["Runic Corruption"] = 19229,
	-- Level 75 talents
	["Death Pact"] = 19226,
	-- Level 90 talents
	["Gorefiend's Grasp"] = 19230,
	["Remorseless Winter"] = 19231,
	["Desecrated Ground"] = 19232,
	-- Level 100 talents
	["Necrotic Plague"] = 21207,
	["Defile"] = 21208,
	["Breath of Sindragosa"] = 21209,
}

-- Procs: Abilities with cooldowns
CompactRunes.AbilityProcsList = {
	["Frost"] = {
		["Pillar of Frost"] = true,
	},
	["Unholy"] = {
		["Summon Gargoyle"] = true,
	},
	["Blood"] = {
	},
	["All"] = {
		["Breath of Sindragosa"] = true,
		["Death and Decay"] = true,
		--["Defile"] = true,
	},
}
CompactRunes.AbilityProcs = {}

CompactRunes.HasRuneCD = {
	["All"] = {
		["Death and Decay"] = true,
		["Defile"] = true,
		["Outbreak"] = true,
		["Strangulate"] = true,
		["Army of the Dead"] = true,
		["Asphyxiate"] = true,
	}
}
CompactRunes.IgnoreRuneCDs = {}

local ItemIds = {
	--["Steadfast Talisman of the Shado-Pan Assault"] = 94507,
	--["Brutal Talisman of the Shado-Pan Assault"] = 94508,
	--["Fortitude of the Zandalari"] = 94516,
	--["Fabled Feather of Ji-Kun"] = 94515,
	--["Thok's Tail Tip"] = 102305,
	--["Fusion-Fire Core"] = 102295,
	--["Evil Eye of Galakras"] = 102298,
	["Scabbard of Kyanos"] = 118882,
	["Pol's Blinded Eye"] = 113834,
	["Vial of Convulsive Shadows"] = 113969,
	["Tablet of Turnbuckle Teamwork"] = 113905,
}
local ItemNames = {}
local ItemIcons = {}
local ReverseItemNames = {}
_G.setmetatable(ItemNames, LookupOrKeyMT)
local function LoadItemNames()
	for k, v in pairs(ItemIds) do
		local name = _G.GetItemInfo(v)
		local icon = _G.GetItemIcon(v)
		ItemNames[k] = name
		ItemIcons[k] = icon
		if name then
			ReverseItemNames[name] = k
		end
	end
end
LoadItemNames()
addon.ItemIds = ItemIds
addon.ItemNames = ItemNames
addon.ItemIcons = ItemIcons
addon.ReverseItemNames = ReverseItemNames

local SpellTypes = {
	["Plague Leech"] = "talent",
	["Unholy Blight"] = "talent",
	["Lichborne"] = "talent",
	["Anti-Magic Zone"] = "talent",
	["Death's Advance"] = "talent",
	["Asphyxiate"] = "talent",
	["Death Pact"] = "talent",
	["Gorefiend's Grasp"] = "talent",
	["Remorseless Winter"] = "talent",
	["Desecrated Ground"] = "talent",
	["Defile"] = "talent",
}
local SpellOverridenByTalent = {
	["Death and Decay"] = "Defile",
	["Strangulate"] = "Asphyxiate",
}

addon.DynamicSpells = {
	["Death and Decay"] = true,
	["Strangulate"] = true,
}
function addon.IsSpellKnown(name)
	local type = SpellTypes[name]
	local overrideTalent = SpellOverridenByTalent[name]
	if overrideTalent and addon.HasActiveTalent(overrideTalent) then
		return false, overrideTalent
	end
	if type and type == "talent" then
		return addon.HasActiveTalent(name)
	else
		return _G.IsSpellKnown(SpellIds[name])
	end
end

local ItemProcs = {
	--[[
	["Fabled Feather of Ji-Kun"] = {
		proc = "Feathers of Fury",
		procIds = { 138759 },
	},
	["Thok's Tail Tip"] = {
		proc = "Determination",
		procIds = { 146250 },
		icd = 115,
	},
	["Fusion-Fire Core"] = {
		proc = "Tenacious",
		procIds = { 148899 },
		icd = 115,
	},
	["Evil Eye of Galakras"] = {
		proc = "Outrage",
		procIds = { 146245 },
		icd = 55,
	},
	["Brutal Talisman of the Shado-Pan Assault"] = {
		proc = "Surge of Strength",
		procIds = { 138702 },
		icd = 85,
	},
]]--
}
addon.ItemProcs = ItemProcs

local Broker = _G.CreateFrame("Frame")
Broker.obj = LDB:NewDataObject(addon.addonTitle, {
	type = "data source",
	icon = "Interface\\Icons\\Spell_DeathKnight_RuneTap",
	label = addon.addonTitle,
	text = addon.addonTitle,
	barValue = 0,
	barR = 0,
	barG = 0,
	barB = 1,
	OnClick = function(clickedframe, button)
		if button == "RightButton" then
			local optionsFrame = _G.InterfaceOptionsFrame

			if optionsFrame:IsVisible() then
				optionsFrame:Hide()
			else
				CompactRunes:ShowOptions()
			end
		elseif button == "LeftButton" and _G.IsShiftKeyDown() then
			CompactRunes.TotalStats:Reset()
			CompactRunes.LastFightStats:Reset()
			addon:FireCallback("ResetStats")		
		end
	end
} )

local GREEN = "|cff00ff00"
local YELLOW = "|cffffff00"
local BLUE = "|cff0198e1"
local ORANGE = "|cffff9933"

local addonHdr = GREEN.."%s %s"
local totalDataHdr = ORANGE..L["Session"]
local lastFightValueHdr = ORANGE..L["Last Fight"]
local reProcFmt = "%s%d|r / %s%d|r / %s%d|r / %s%d"
local reProcTypeHdr = "    %s%s"

local valuePlusPercent = addon.valuePlusPercent

local kmProcTitle, kmOverwrittenTitle, kmExpiredTitle, kmFSTitle, kmOblitTitle
local function AddStats(tooltip, statstype)
	local stats
	if statstype == "LastFight" then
		stats = CompactRunes.LastFightStats
	else
		stats = CompactRunes.TotalStats
	end
	local totalProcs = 0
	local reProcs = stats.reProcs
	for k, v in pairs(reProcs) do
		totalProcs = totalProcs + (v or 0)
	end
	local reRegenUses = stats.regenUses["RE"] or 0
	local rcRegenUses = stats.regenUses["RC"] or 0
	local btRegenUses = stats.regenUses["BT"] or 0

	tooltip:AddSeparator(1)
	tooltip:AddLine(L["Blood Charges Capped"],
		valuePlusPercent(stats.cappedBC, btRegenUses))
	tooltip:AddLine(L["Runic Corruption Procs"],
		valuePlusPercent(stats.rcProcs, rcRegenUses))
	tooltip:AddLine(L["Runic Empowerment Procs"],
		valuePlusPercent(totalProcs, reRegenUses))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["BLOOD"],"Blood"),
			valuePlusPercent(reProcs[1], totalProcs))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["UNHOLY"],"Unholy"),
			valuePlusPercent(reProcs[2], totalProcs))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["FROST"],"Frost"),
			valuePlusPercent(reProcs[3], totalProcs))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["DEATH"],"Death"),
			valuePlusPercent(reProcs[4], totalProcs))

	for name, obj in pairs(addon.modules) do
		if obj and obj.AddStats then
			obj:AddStats(tooltip, statstype)
		end
	end
end

function Broker.obj:OnEnter()
	local tooltip = LibQTip:Acquire("CompactRunesTooltip", 2, "LEFT", "RIGHT")
	self.tooltip = tooltip 

	tooltip:AddHeader(addonHdr:format(addon.addonTitle, addon.addonVersion))
	tooltip:AddLine()

	if CompactRunes.isDK then
		tooltip:AddLine(L["Shift + Left-Click to reset"], "", 1, 1, 1)
		tooltip:AddLine()

		tooltip:AddLine(totalDataHdr)
		AddStats(tooltip, "Session")

		tooltip:AddLine()
		tooltip:AddLine(lastFightValueHdr)
		AddStats(tooltip, "LastFight")	
	end

	tooltip:SmartAnchorTo(self)
	tooltip:Show()
end

function Broker.obj:OnLeave()
	LibQTip:Release(self.tooltip)
	self.tooltip = nil
end

local frames = {}
addon.frames = frames

CompactRunes.PriorityListDefaults = {
	["Blood"] = {
		"Soul Reaper",
		"Blood Charge (10+)",
		"Plague Leech",
		"Crimson Scourge",
		"Dark Succor",
		"Blood Charge (5-9)",
		"Blood Charge",
		"Breath of Sindragosa",
		"Death and Decay",
	},
	["Frost"] = {
		"Blood Charge (10+)",
		"Soul Reaper",
		"Plague Leech",
		"Killing Machine", 
		"Freezing Fog", 
		"Blood Charge (5-9)",
		"Blood Charge",
		"Dark Succor",
		"Pillar of Frost",
		"Breath of Sindragosa",
		"Death and Decay",
	},
	["Unholy"] = {
		"Raise Dead",
		"Dark Transformation",
		"Soul Reaper",
		"Blood Charge (10+)",
		"Plague Leech",
		"Sudden Doom", 
		"Summon Gargoyle",
		"Blood Charge (5-9)",
		"Blood Charge",
		"Dark Succor",
		"Breath of Sindragosa",
		"Death and Decay",
	},
}

CompactRunes.DefaultCDBarOrderings = {
	["Blood"] = {
		"Rune Tap",
		"Vampiric Blood",
		"Anti-Magic Shell",
		"Outbreak",
		"Dancing Rune Weapon",
		"Death Pact",
		"Icebound Fortitude",
		"Empower Rune Weapon",
		"Army of the Dead",
		"Death and Decay",
		"Gorefiend's Grasp",
		"Remorseless Winter",
		"Desecrated Ground",
		"Defile",
		"Breath of Sindragosa",
		"Plague Leech",
		"Unholy Blight",
		"Lichborne",
		"Anti-Magic Zone",
		"Death's Advance",
		"Mind Freeze",
		"Strangulate",
		"Asphyxiate",
		--"Scabbard of Kyanos",
		"Pol's Blinded Eye",
		"Tablet of Turnbuckle Teamwork",
		"Vial of Convulsive Shadows",
	},
	["Frost"] = {
		"Pillar of Frost",
		"Outbreak",
		"Anti-Magic Shell",
		"Empower Rune Weapon",
		"Army of the Dead",
		"Death and Decay",
		"Gorefiend's Grasp",
		"Remorseless Winter",
		"Desecrated Ground",
		"Defile",
		"Breath of Sindragosa",
		"Plague Leech",
		"Unholy Blight",
		"Lichborne",
		"Anti-Magic Zone",
		"Death's Advance",
		"Mind Freeze",
		"Strangulate",
		"Asphyxiate",
		"Icebound Fortitude",
		"Death Pact",
		"Vial of Convulsive Shadows",
		"Scabbard of Kyanos",
	},
	["Unholy"] = {
		"Summon Gargoyle",
		"Anti-Magic Shell",
		"Outbreak",
		"Empower Rune Weapon",
		"Army of the Dead",
		"Death and Decay",
		"Gorefiend's Grasp",
		"Remorseless Winter",
		"Desecrated Ground",
		"Defile",
		"Breath of Sindragosa",
		"Plague Leech",
		"Unholy Blight",
		"Lichborne",
		"Anti-Magic Zone",
		"Death's Advance",
		"Mind Freeze",
		"Strangulate",
		"Asphyxiate",
		"Icebound Fortitude",
		"Death Pact",
		"Vial of Convulsive Shadows",
		"Scabbard of Kyanos",
	},
}

-- These procs are handled differently.
CompactRunes.specialProcs = {
	["Soul Reaper"] = true,
	["Plague Leech"] = true,
	["Pillar of Frost"] = true,
	["Summon Gargoyle"] = true,
	["Raise Dead"] = true,
	["Blood Charge (10+)"] = true,
	["Blood Charge (5-9)"] = true,
	["Death and Decay"] = true,
	["Defile"] = true,
	["Breath of Sindragosa"] = true,
}

CompactRunes.soulReaperIds = {
	["Blood"] = SpellIds["Soul Reaper (Blood)"],
	["Frost"] = SpellIds["Soul Reaper (Frost)"],
	["Unholy"] = SpellIds["Soul Reaper (Unholy)"]
}

CompactRunes.regenAbilities = {
	[SpellIds["Death Coil"]] = true,
	[SpellIds["Rune Strike"]] = true,
	[SpellIds["Frost Strike"]] = true,
	[SpellIds["Death Coil Damage"]] = true,
}

function CompactRunes:MergeLists(list, settings, defaults)
	local uniques = {}
	local added = {}
	for i, v in ipairs(defaults) do
		uniques[v] = true
	end
	if settings then
		for i, v in ipairs(settings) do
			if v and uniques[v] and not added[v] then
				tinsert(list, v)
				added[v] = true
			end
		end
	end
	for i, v in ipairs(defaults) do
		if v and not added[v] then
			tinsert(list, v)
			added[v] = true
		end
	end
end

function CompactRunes:UpdatePriorityListSettings()
	for spec, list in pairs(self.PriorityListDefaults) do
		local list = {}
		self:MergeLists(
			list,
			self.db.profile.priorityList[spec],
			self.PriorityListDefaults[spec])
		self.db.profile.priorityList[spec] = list
	end
end

function CompactRunes:UpdateCDBarOrderings()
	for spec, list in pairs(self.DefaultCDBarOrderings) do
		local list = {}
		self:MergeLists(
			list,
			self.db.profile.cdbars[spec].ordering,
			self.DefaultCDBarOrderings[spec])
		self.db.profile.cdbars[spec].ordering = list
	end
end

function CompactRunes:MigrateSettings()
	local version = self.db.profile.profileVersion
	if version and version < 2 then
		for spec, list in pairs(self.db.profile.priorityList) do
			self.db.profile.priorityList[spec] = {}
		end
	end
	self.db.profile.profileVersion = 2

	if self.db.profile.runeRechargeNums ~= nil then
		if self.db.profile.runeRechargeNums then
			self.db.profile.rechargeTimers = "All"
		end
		self.db.profile.runeRechargeNums = nil
	end
	if self.db.profile.width ~= nil then
		self.db.profile.dimensions["HORIZONTAL"].width = self.db.profile.width
		self.db.profile.width = nil
	end
	if self.db.profile.height ~= nil then
		self.db.profile.dimensions["HORIZONTAL"].height = self.db.profile.height
		self.db.profile.height = nil
	end
	if self.db.profile.spacing ~= nil then
		self.db.profile.dimensions["HORIZONTAL"].spacing = self.db.profile.spacing
		self.db.profile.spacing = nil
	end
	if self.db.profile.runes and self.db.profile.runes.inset ~= nil then
		self.db.profile.runes.inset = nil
		self.db.profile.runes = nil
	end
end

function CompactRunes:ResetPriorityList(spec)
	self.db.profile.priorityList[spec] = {}
	self:UpdatePriorityList(spec)
end

function CompactRunes:RemoveDefaults()
	for spec, list in pairs(self.PriorityListDefaults) do
		local matches = true
		local profile = self.db.profile.priorityList[spec]
		for i, v in ipairs(list) do
			if not (v and profile[i] and profile[i] == v) then
				matches = false
			end
		end
		if matches then
			self.db.profile.priorityList[spec] = {}
		end
	end
end

addon.defaults = {
	profile = {
		minimap = {
			hide = true,
		},
		debug = false,
		debugBP = false,
		locked = false,
		hideBlizzardFrames = false,
		ooc_alpha = 0.4,
		combatAlpha = 1.0,
		oocRechargeAlphaEnabled = false,
		oocRechargeAlpha = 0.4,
		rechargeAlpha = 0.3,
		showCDText = true,
		rechargeTimers = "None",
		runeOrientation = "HORIZONTAL",
		almostReadyTime = 0,
		almostReadyAlpha = 0.6,
		runeBackground = 1,
		runeBackgroundAlpha = 0.1,
		runeBackgrounds = {
			transparent = { r = 0.0, g = 0.0, b = 0.0, a = 0.0 },
			black = { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
		},
		RuneColors = {
			["BLOOD"] = { r = 0.77, g = 0.12, b = 0.23, a = 1.0 },
			["UNHOLY"] = { r = 0.3, g = 0.8, b = 0.1, a = 1.0 },
			["FROST"] = { r = 0.0, g = 0.4, b = 0.7, a = 1.0 },
			["DEATH"] = { r = 0.51, g = 0.23, b = 0.65, a = 1.0 },
		},
		RuneCooldownColors = {
			["BLOOD"] = { r = 1.0, g = 0.0, b = 0.0, a = 0.1 },
			["UNHOLY"] = { r = 0.0, g = 1.0, b = 0.0, a = 0.1 },
			["FROST"] = { r = 0.0, g = 0.0, b = 1.0, a = 0.1 },
			["DEATH"] = { r = 1.0, g = 0.0, b = 1.0, a = 0.1 },
			["NONE"] = { r = 0.0, g = 0.0, b = 0.0, a = 0.0 },
		},
		rangeInd = true,
		rangeColor = { r = 1.0, g = 0.0, b = 0.0, a = 0.6 },
		rcProcInd = true,
		rcProcColor = { r = 1.0, g = 1.0, b = 1.0, a = 0.6 },
		plRefresh = 3,
		plOutbreak = true,
		plRequireRunes = true,
		kmStats = false,
		diseaseThresholdType = "Percent",
		diseaseThreshold = 0.1,
		diseaseThresholdAmt = 1000,
		font_face = "Friz Quadrata TT",
		font_size = 12,
		fontFlags = {
			OUTLINE = true,
			THICKOUTLINE = false,
			MONOCHROME = false
		},
		texture = "Compact Runes",
		border = "Compact Runes",
		borderInd = "Compact Runes Bright",
		runeOrder = "BFU",
		diseaseOrder = "BF",
		dimensions = {
			["HORIZONTAL"] = {
				width = 30,
				height = 10,
				spacing = 4,
				inset = 6,
				swap = false,
				barSpacing = 6,
			},
			["VERTICAL"] = {
				width = 12,
				height = 25,
				spacing = 4,
				inset = 6,
				swap = true,
				barSpacing = 6,
			},
		},
		["frame"] = {
			x = 0,
			y = 0,
			color = { r = 0.1, g = 0.1, b = 0.1, a = 1.0 },
			bdcolor = { r = 0.5, g = 0.5, b = 0.5, a = 1.0 },
			rccolor = { r = 1.0, g = 1.0, b = 1.0, a = 0.7 },
		},
		rpTicks = {
			enabled = false,
			color = { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
			width = 2,
			specs = {
				["Blood"] = {
					cost = 30,
				},
				["Frost"] = {
					cost = 25,
				},
				["Unholy"] = {
					cost = 30,
				},
			},
		},
		["procIcon"] = {
			scale = 1.0,
			userPlaced = false,
			x = 0,
			y = 0,
			alphas = {
				combat = 1.0,
				ooc = 0.6,
			},
		},
		["srcdIcon"] = {
			enabled = true,
			scale = 1.0,
			userPlaced = false,
			x = 0,
			y = 0,
			alphas = {
				combat = 0.2,
				ooc = 0.1,
				active = 1.0,
			},
		},
		procs = true,
		ignoreRuneCDs = {
			["Soul Reaper"] = true,
			["Procs"] = true,
			["CooldownBar"] = true,
		},
		enabled = {
			["Blood"] = {
				["Blood Charge (10+)"] = false,
				["Blood Charge (5-9)"] = false,
				["Blood Charge"] = false,
				["Soul Reaper"] = true,
				["Crimson Scourge"] = true,					
				["Dark Succor"] = true,
				["Plague Leech"] = false,
				["Breath of Sindragosa"] = false,
				["Death and Decay"] = false,
			},
			["Frost"] = {
				["Soul Reaper"] = true,
				["Freezing Fog"] = true, 
				["Killing Machine"] = true, 
				["Blood Charge (10+)"] = false,
				["Blood Charge (5-9)"] = false,
				["Blood Charge"] = false,
				["Dark Succor"] = true,
				["Plague Leech"] = false,
				["Breath of Sindragosa"] = false,
				["Death and Decay"] = false,
			},
			["Unholy"] = {
				["Raise Dead"] = true,
				["Soul Reaper"] = true,
				["Sudden Doom"] = true, 
				["Blood Charge (10+)"] = false,
				["Blood Charge (5-9)"] = false,
				["Blood Charge"] = false,
				["Dark Succor"] = true,
				["Plague Leech"] = false,
				["Dark Transformation"] = false,
				["Breath of Sindragosa"] = false,
				["Death and Decay"] = false,
			},
		},
		priorityList = {
			["Blood"] = {},
			["Frost"] = {},
			["Unholy"] = {},
		},
		cdbars = {
			['**'] = {
				enabled = false,
				x = nil,
				y = nil,
				spacing = 5,
				maxItemsPerRow = 10,
				maxRows = 2,
				oocAlpha = 0.6,
				combatAlpha = 1.0,
				inactiveAlpha = 0.4,
				activeAlpha = 1.0,
				textcolor = { r = 0.85, g = 0.85, b = 0.0, a = 1.0 },
				textFontSize = 14,
				fontFace = "Friz Quadrata TT",
				height = 32,
				width = 32,
			},
			["Unholy"] = {
				cds = {
					["Outbreak"] = {
						enabled = true,
						type = "spell",
					},
					["Summon Gargoyle"] = {
						enabled = true,
						type = "spell",
					},
					["Anti-Magic Shell"] = {
						enabled = true,
						type = "spell",
					},
					["Empower Rune Weapon"] = {
						enabled = true,
						type = "spell",
					},
					["Army of the Dead"] = {
						enabled = true,
						type = "spell",
					},
					["Death and Decay"] = {
						enabled = true,
						type = "spell",
					},
					["Gorefiend's Grasp"] = {
						enabled = true,
						type = "spell",
					},
					["Remorseless Winter"] = {
						enabled = true,
						type = "spell",
					},
					["Desecrated Ground"] = {
						enabled = true,
						type = "spell",
					},
					["Defile"] = {
						enabled = true,
						type = "spell",
					},
					["Breath of Sindragosa"] = {
						enabled = true,
						type = "spell",
					},
					["Plague Leech"] = {
						enabled = false,
						type = "spell",
					},
					["Unholy Blight"] = {
						enabled = true,
						type = "spell",
					},
					["Lichborne"] = {
						enabled = true,
						type = "spell",
					},
					["Death's Advance"] = {
						enabled = true,
						type = "spell",
					},
					["Anti-Magic Zone"] = {
						enabled = true,
						type = "spell",
					},
					["Mind Freeze"] = {
						enabled = false,
						type = "spell",
					},
					["Strangulate"] = {
						enabled = true,
						type = "spell",
					},
					["Asphyxiate"] = {
						enabled = true,
						type = "spell",
					},
					["Icebound Fortitude"] = {
						enabled = true,
						type = "spell",
					},
					["Death Pact"] = {
						enabled = true,
						type = "spell",
					},
					-- Items
					["Scabbard of Kyanos"] = {
						enabled = true,
						type = "item",
					},
					["Vial of Convulsive Shadows"] = {
						enabled = true,
						type = "item",
					},
				},
				ordering = {},
			},
			["Frost"] = {
				cds = {
					["Outbreak"] = {
						enabled = true,
						type = "spell",
					},
					["Pillar of Frost"] = {
						enabled = true,
						type = "spell",
					},
					["Anti-Magic Shell"] = {
						enabled = true,
						type = "spell",
					},
					["Empower Rune Weapon"] = {
						enabled = true,
						type = "spell",
					},
					["Army of the Dead"] = {
						enabled = true,
						type = "spell",
					},
					["Death and Decay"] = {
						enabled = true,
						type = "spell",
					},
					["Gorefiend's Grasp"] = {
						enabled = true,
						type = "spell",
					},
					["Remorseless Winter"] = {
						enabled = true,
						type = "spell",
					},
					["Desecrated Ground"] = {
						enabled = true,
						type = "spell",
					},
					["Defile"] = {
						enabled = true,
						type = "spell",
					},
					["Breath of Sindragosa"] = {
						enabled = true,
						type = "spell",
					},
					["Plague Leech"] = {
						enabled = false,
						type = "spell",
					},
					["Unholy Blight"] = {
						enabled = true,
						type = "spell",
					},
					["Lichborne"] = {
						enabled = true,
						type = "spell",
					},
					["Death's Advance"] = {
						enabled = true,
						type = "spell",
					},
					["Anti-Magic Zone"] = {
						enabled = true,
						type = "spell",
					},
					["Mind Freeze"] = {
						enabled = false,
						type = "spell",
					},
					["Strangulate"] = {
						enabled = true,
						type = "spell",
					},
					["Asphyxiate"] = {
						enabled = true,
						type = "spell",
					},
					["Icebound Fortitude"] = {
						enabled = true,
						type = "spell",
					},
					["Death Pact"] = {
						enabled = true,
						type = "spell",
					},
					-- Items
					["Scabbard of Kyanos"] = {
						enabled = true,
						type = "item",
					},
					["Vial of Convulsive Shadows"] = {
						enabled = true,
						type = "item",
					},
				},
				ordering = {},
			},
			["Blood"] = {
				cds = {
					["Outbreak"] = {
						enabled = true,
						type = "spell",
					},
					["Rune Tap"] = {
						enabled = true,
						type = "spellcharges",
					},
					["Vampiric Blood"] = {
						enabled = true,
						type = "spell",
					},
					["Anti-Magic Shell"] = {
						enabled = true,
						type = "spell",
					},
					["Dancing Rune Weapon"] = {
						enabled = true,
						type = "spell",
					},
					["Death Pact"] = {
						enabled = true,
						type = "spell",
					},
					["Icebound Fortitude"] = {
						enabled = true,
						type = "spell",
					},
					["Army of the Dead"] = {
						enabled = true,
						type = "spell",
					},
					["Empower Rune Weapon"] = {
						enabled = true,
						type = "spell",
					},
					["Death and Decay"] = {
						enabled = true,
						type = "spell",
					},
					["Gorefiend's Grasp"] = {
						enabled = true,
						type = "spell",
					},
					["Remorseless Winter"] = {
						enabled = true,
						type = "spell",
					},
					["Desecrated Ground"] = {
						enabled = true,
						type = "spell",
					},
					["Defile"] = {
						enabled = true,
						type = "spell",
					},
					["Breath of Sindragosa"] = {
						enabled = true,
						type = "spell",
					},
					["Plague Leech"] = {
						enabled = false,
						type = "spell",
					},
					["Unholy Blight"] = {
						enabled = true,
						type = "spell",
					},
					["Lichborne"] = {
						enabled = true,
						type = "spell",
					},
					["Death's Advance"] = {
						enabled = true,
						type = "spell",
					},
					["Anti-Magic Zone"] = {
						enabled = true,
						type = "spell",
					},
					["Mind Freeze"] = {
						enabled = false,
						type = "spell",
					},
					["Strangulate"] = {
						enabled = true,
						type = "spell",
					},
					["Asphyxiate"] = {
						enabled = true,
						type = "spell",
					},
					-- Items
					["Pol's Blinded Eye"] = {
						enabled = true,
						type = "item",
					},
					["Tablet of Turnbuckle Teamwork"] = {
						enabled = true,
						type = "item",
					},
					["Vial of Convulsive Shadows"] = {
						enabled = true,
						type = "item",
					},
				},
				ordering = {},
			},
		},
		blocks = {
			['**'] = {
				width = 30,
				height = 10,
				textcolor = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
				userPlaced = false,
				x = nil,
				y = nil,
			},
			["NecroticBlight"] = {
				enabled = false,
				width = 30,
				height = 30,
				userPlaced = true,
				x = 50,
				y = -30,
				fontSize = 16,
				threshold = 15,
				combatAlpha = 0.5,
				oocAlpha = 0.0,
				diseasedAlpha = 0.7,
				alwaysShowCount = false,
				indicators = {
					["Low"] = { r = 1.0, g = 0.0, b = 0.0, a = 1.0 },
					["Good"] = { r = 0.0, g = 1.0, b = 0.0, a = 1.0 },
				},
			},
			["T18Frost2POblit"] = {
				enabled = false,
				width = 30,
				height = 30,
				userPlaced = true,
				x = -80,
				y = -30,
				fontSize = 16,
				combatAlpha = 0.0,
				oocAlpha = 0.0,
				activeAlpha = 1.0,
				textcolor = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
			},
			["T18Frost2PFW"] = {
				enabled = false,
				width = 30,
				height = 30,
				userPlaced = true,
				x = -50,
				y = -30,
				fontSize = 16,
				combatAlpha = 0.0,
				oocAlpha = 0.0,
				activeAlpha = 1.0,
				textcolor = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
			},
		},
		-- Bars
		bars = {
			-- Provide defaults for all bars.
			-- These are inherited if no bar or no value is set.
			['**'] = {
				textcolor = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
				userPlaced = false,
				x = nil,
				y = nil,
				overrideHeight = false,
				overrideWidth = false,
				height = nil,
				width = nil,
			},
			["RunicPowerBar"] = {
				enabled = true,
				color = { r = 0.0, g = 0.82, b = 1.0, a = 1.0 },
				bgcolor = { r = 0.0, g = 0.205, b = 0.25, a = 1.0 },
				textcolor = { r = 0.66, g = 0.82, b = 1.0, a = 1.0 },
				showNumericValue = true,
			},
			["BloodShieldBar"] = {
				enabled = false,
				color = { r = 1.0, g = 0.1, b = 0.1, a = 1.0 },
				bgcolor = { r = 0.5, g = 0.05, b = 0.05, a = 1.0 },
				textcolor = { r = 1.0, g = 0.85, b = 0.85, a = 1.0 },
			},
			["FrostFeverBar"] = {
				enabled = true,
				color = { r = 0.0, g = 0.45, b = 0.6, a = 1.0 },
				bgcolor = { r = 0.0, g = 0.1125, b = 0.15, a = 0.9 },
				textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
			},
			["BloodPlagueBar"] = {
				enabled = true,
				color = { r = 0.5, g = 0.0, b = 0.6, a = 1.0 },
				bgcolor = { r = 0.125, g = 0.0, b = 0.15, a = 0.9 },
				textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
			},
			["DarkTransformationBar"] = {
				enabled = true,
				color = { r = 0.0, g = 0.5, b = 0.0, a = 1.0 },
				bgcolor = { r = 0.125, g = 0.0, b = 0.15, a = 0.9 },
				textcolor = { r = 0.83, g = 0.66, b = 0.86, a = 1.0 },
				alt_color = { r = 0.5, g = 0.25, b = 0.0, a = 1.0 },
				alt_textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
			},
			["BloodChargeBar"] = {
				enabled = true,
				color = { r = 1.0, g = 0.0, b = 0.0, a = 1 },
				bgcolor = { r = 0.65, g = 0.0, b = 0.0, a = 0.8 },
				textcolor = { r = 1.0, g = 0.9, b = 0.9, a = 1.0 },
			},
			["BoneShieldBar"] = {
				enabled = false,
				progress = "Time",
				showCooldown = true,
				showReady = true,
				color = { r = 0.1, g = 0.45, b = 0.1, a = 1.0 },
				bgcolor = { r = 0.05, g = 0.15, b = 0.05, a = 0.9 },
				textcolor = { r = 0.9, g = 1.0, b = 0.9, a = 1.0 },
				rechargeTextColor = { r = 1.0, g = 0.7, b = 0.7, a = 1.0 },
			},
			["NecroticPlagueBar"] = {
				enabled = true,
				color = { r = 0.6, g = 0.5, b = 0.1, a = 1.0 },
				bgcolor = { r = 0.15, g = 0.125, b = 0.0, a = 0.9 },
				textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
				secondaryColor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
				progress = "Time",
				primary = "Time",
				secondaryPos = "LEFT",
			},
			["SwingTimerBar"] = {
				enabled = false,
				color = { r = 0.7, g = 0.7, b = 0.7, a = 1.0 },
				bgcolor = { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
				userPlaced = true,
				x = 0,
				y = -30,
				numericTimer = true,
				specs = {
					["Blood"] = false,
					["Frost"] = true,
					["Unholy"] = false,
				},
			},

--[[
			["GargoyleBar"] = {
				enabled = false,
				progress = "Time",
				showCooldown = true,
				showReady = true,
				color = { r = 0.1, g = 0.45, b = 0.1, a = 1.0 },
				bgcolor = { r = 0.05, g = 0.15, b = 0.05, a = 0.9 },
				textcolor = { r = 0.9, g = 1.0, b = 0.9, a = 1.0 },
				rechargeTextColor = { r = 1.0, g = 0.7, b = 0.7, a = 1.0 },
			},
]]--
		},
		events = {
			["**"] = {
				sound = "None",
				soundChannel = "Master",
			},
			["Blood Charge"] = {
				enabled = false,
				count = 10,
			},
		},
	},
}

function CompactRunes:SetRuneBackground()
	local setting = self.db.profile.runeBackground

	if setting == nil or tonumber(setting) == nil or setting < 1 or setting > 3 then
		self.db.profile.runeBackground = 1
		setting = 1
	end

	if setting == 3 then
		self.altRuneBkg = self.db.profile.runeBackgrounds.black
	else
		self.altRuneBkg = self.db.profile.runeBackgrounds.transparent
	end
end

function CompactRunes:UpdateBoneShieldBarMode()
	local bar = frames.boneBar
	if bar.db.progress == "Time" then
		bar.bar:SetMinMaxValues(0, 1)
		bar.bar:SetValue(1)
	elseif bar.db.progress == "Charges" then
		bar.bar:SetMinMaxValues(0, addon.MAX_BONESHIELD_CHARGES)
		bar.bar:SetValue(0)
	elseif bar.db.progress == "None" then
		bar.bar:SetMinMaxValues(0, 1)
		bar.bar:SetValue(1)        
	end
end

function CompactRunes:OnInitialize()
	self:CheckClass()
	if not self.isDK then return end

	-- Load the settings
	self.db = _G.LibStub("AceDB-3.0"):New(
		"CompactRunesDB", addon.defaults, "Default")
	addon.profile = self.db.profile

	self:MigrateSettings()
	self:UpdateRuneColors()
	self:UpdatePriorityListSettings()
	self:UpdateCDBarOrderings()

	-- Register for profile callbacks
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChange")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChange")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChange")
	self.db.RegisterCallback(self, "OnProfileShutdown", "RemoveDefaults")
	self.db.RegisterCallback(self, "OnDatabaseShutdown", "RemoveDefaults")
	
	LSM:Register("statusbar", "Compact Runes",
		[[Interface\Addons\CompactRunes\CompactRunes]])
	LSM:Register("border", "Compact Runes",
		[[Interface\Addons\CompactRunes\CompactRunesBorder]])
	LSM:Register("border", "Compact Runes Bright",
		[[Interface\Addons\CompactRunes\CompactRunesBrightBorder]])

	Icon:Register("CompactRunesLDB", Broker.obj, self.db.profile.minimap)
	LSM.RegisterCallback(CompactRunes, "LibSharedMedia_Registered")

	for name, obj in pairs(addon.modules) do
		if obj and obj.OnInitialize then
			obj:OnInitialize()
		end
	end
end

function CompactRunes:OnProfileChange()
	self:MigrateSettings()
	self:UpdateRuneColors()
	self:UpdatePriorityListSettings()
	self:UpdateCDBarOrderings()
	self:Reset()
	addon.profile = self.db.profile
	addon:FireCallback("ProfileUpdate")
end

function CompactRunes:Reset()
	--self:ResetFonts()
	--self:UpdateTextures()

	self:UpdateRuneColors()
	self:SetRuneBackground()

	for k, v in pairs(self.bars) do
		if v then
			v.db = self.db.profile.bars[v.name]
			v:Reset()
		end
	end

	self:SetRuneOrder(self.db.profile.runeOrder)

	local procInd = frames.procInd
	if procInd then
		procInd.ResetFonts()
		procInd:SetWidth(procInd:GetDesiredWidth())
		procInd:SetHeight(procInd:GetDesiredHeight())
		procInd:SetDesiredPoint()
	end

	self:SetMainFrameBorder()
	local main = frames.main
	if main then
		main:SetWidth(main.GetDesiredWidth())
		main:SetHeight(main.GetDesiredHeight())
	end
end

function CompactRunes:LibSharedMedia_Registered(event, mediatype, key)
	if _G.strlen(self.db.profile.font_face) > 1 and mediatype == "font" then
		if self.db.profile.font_face == key then
			self:ResetFonts()
		end
	end
	if mediatype == "statusbar" then
	    self:UpdateTextures()
	end
end

function CompactRunes:ResetFonts()
	for k, v in pairs(self.bars) do
		if v then
			v:ResetFonts()
		end
	end

	if frames.procInd then
		frames.procInd.ResetFonts()
	end
end

function CompactRunes:UpdateTextures()
	for k, v in pairs(self.bars) do
		if v then
			v:UpdateTexture()
		end
	end
end

local function splitWords(str)
  local w = {}
  local function helper(word) table.insert(w, word) return nil end
  str:gsub("(%w+)", helper)
  return w
end

function CompactRunes:ChatCommand(input)
	if not input or input:trim() == "" then
		self:ShowOptions()
	else
		local cmds = splitWords(input)
		if cmds[1] and cmds[1] == "debug" then
			if cmds[2] and cmds[2] == "on" then
				self.db.profile.debug = true
				self:Print("Debugging on.  Use '/crunes debug off' to disable.")
			elseif cmds[2] and cmds[2] == "off" then
				self.db.profile.debug = false
				self:Print("Debugging off.")
			else
				self:Print("Debugging is "..(self.db.profile.debug and "on." or "off."))
			end
		elseif cmds[1] and cmds[1] == "srMethod" then
			if cmds[2] and cmds[2] == "cooldown" then
				self.srMethod = "cooldown"
			elseif cmds[2] and cmds[2] == "target" then
				self.srMethod = "target"
			end
			self:Print("Soul Reaper detected via " ..
				tostring(self.srMethod))
		end
	end
end

function CompactRunes:OnEnable()
	if not self.isDK then return end

	self:CheckTalents()
	self:SetRuneBackground()
	self:Load()
	self:CheckGear()
	LoadSpellNames()
	LoadItemNames()
	addon.SetupSpellAuras()
	self:HideBlizzardFrames()

	if not self.optionsFrame then
		-- Register Options
		local displayName = addon.addonTitle
		self:GetOptions()
		_G.LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable(
			displayName, addon.options)

		self.optionsFrame = {}
		local ACD = _G.LibStub("AceConfigDialog-3.0")
		self.optionsFrame.Main = ACD:AddToBlizOptions(
			displayName, displayName, nil, "core")
		self.optionsFrame.Layout = ACD:AddToBlizOptions(
			displayName, L["Layout"], displayName, "layout")
		self.optionsFrame.Colors = ACD:AddToBlizOptions(
			displayName, L["Colors"], displayName, "colors")
		self.optionsFrame.Procs = ACD:AddToBlizOptions(
			displayName, L["Procs"], displayName, "procs")
		self.optionsFrame.Sounds = ACD:AddToBlizOptions(
			displayName, L["Sounds"], displayName, "sounds")
		--self.optionsFrame.IgnoreRunes = ACD:AddToBlizOptions(
		--	displayName, L["Ignore Rune CDs"], displayName, "ignoreRunes")
		self.optionsFrame.CDBars = ACD:AddToBlizOptions(
			displayName, L["Cooldown Bars"], displayName, "cooldowns")

		-- Add options for modules
		for name, obj in pairs(addon.modules) do
			if obj and obj.AddOptions then
				local name, sectionName, tableName = obj.AddOptions()
				self.optionsFrame[name] = ACD:AddToBlizOptions(
					displayName, sectionName, displayName, tableName)
			end
		end

		ACD:AddToBlizOptions(
			displayName, addon.options.args.profile.name, displayName, "profile")
		-- Register the chat command
		self:RegisterChatCommand("crunes", "ChatCommand")
	end

	for name, obj in pairs(addon.modules) do
		if obj and obj.Enable then
			obj:Enable()
		end
	end
end

function CompactRunes:HideBlizzardFrames()
	if not self.db.profile.hideBlizzardFrames then return end
	-- Securely and safely hide and then hook the OnShow to hide itself.
	-- It should last until the UI is reloaded.
	local RuneFrame = _G.RuneFrame
	if RuneFrame and RuneFrame.Hide and _G.type(RuneFrame.Hide) == "function" then
		RuneFrame:Hide()
		if RuneFrame.HookScript and _G.type(RuneFrame.HookScript) == "function" then
			RuneFrame:HookScript("OnShow", function(self) self:Hide() end)
		end
	end
end

local WatchedEvents = {
	"RUNE_POWER_UPDATE",
	"RUNE_TYPE_UPDATE",
	"PLAYER_REGEN_DISABLED",
	"PLAYER_REGEN_ENABLED",
	"PLAYER_TARGET_CHANGED",
	"PLAYER_TALENT_UPDATE",
	"PLAYER_SPECIALIZATION_CHANGED",
	"COMBAT_LOG_EVENT_UNFILTERED",
	"PLAYER_ENTERING_WORLD",
	"UNIT_LEVEL",
	"SPELL_UPDATE_COOLDOWN",
	"PLAYER_EQUIPMENT_CHANGED",
	"PLAYER_ALIVE",
	"PLAYER_DEAD",
}

local UnitEvents = {
	["player"] = {
		"UNIT_POWER_FREQUENT", 
		"UNIT_MAXPOWER",
		"UNIT_AURA",
		"UNIT_PET",
		--"UNIT_SPELLCAST_SUCCEEDED",
	},
	["target"] = {
		"UNIT_HEALTH",
		"UNIT_MAXHEALTH",
		"UNIT_AURA",
	},
	["pet"] = {
		"UNIT_AURA",
	},
}
local function EventFrame_OnEvent(frame, event, ...)
	if event == "UNIT_ATTACK_POWER" then
		CompactRunes:UNIT_ATTACK_POWER(event, ...)
	elseif event == "UNIT_STATS" then
		CompactRunes:UNIT_STATS(event, ...)
	elseif event == "UNIT_POWER_FREQUENT" then
		CompactRunes:UNIT_POWER_FREQUENT(event, ...)
	elseif event == "UNIT_MAXPOWER" then
		CompactRunes:UNIT_MAXPOWER(event, ...)
	elseif event == "UNIT_HEALTH" then
		CompactRunes:UNIT_HEALTH(event, ...)
	elseif event == "UNIT_MAXHEALTH" then
		CompactRunes:UNIT_MAXHEALTH(event, ...)
	elseif event == "UNIT_AURA" then
		CompactRunes:UNIT_AURA(event, ...)
	elseif event == "UNIT_SPELLCAST_SUCCEEDED" then
		CompactRunes:UNIT_SPELLCAST_SUCCEEDED(event, ...)
	elseif event == "UNIT_PET" then
		CompactRunes:UNIT_PET(event, ...)
	elseif event == "PLAYER_ALIVE" then
		CompactRunes:PLAYER_ALIVE(event, ...)
	elseif event == "PLAYER_DEAD" then
		CompactRunes:PLAYER_DEAD(event, ...)
	end
end
local EventFrames = {}
for unit, events in pairs(UnitEvents) do
	local frame = _G.CreateFrame("Frame", ADDON_NAME.."_EventFrame_"..unit)
	frame:SetScript("OnEvent", EventFrame_OnEvent)
	EventFrames[unit] = frame
end

function CompactRunes:Load()
	if self.isDK then
		self:CreateDisplay()
		frames.powerBar:UpdateDisplay()
		for i = 1, #WatchedEvents do
			self:RegisterEvent(WatchedEvents[i])
		end
		for unit, events in pairs(UnitEvents) do
			local eventFrame = EventFrames[unit]
			if eventFrame then
				for i, event in ipairs(events) do
					eventFrame:RegisterUnitEvent(event, unit)
				end
			else
				self:Print("Missing event frame for "..tostring(unit).."!")
			end
		end
	end
end

function CompactRunes:Unload()
end

function CompactRunes:OnDisable()
	for i = 1, #WatchedEvents do
		self:UnregisterEvent(WatchedEvents[i])
	end
	for unit, events in pairs(UnitEvents) do
		local eventFrame = EventFrames[unit]
		if eventFrame then
			for i, event in ipairs(events) do
				eventFrame:UnregisterEvent(event, unit)
			end
		end
	end
	for name, obj in pairs(addon.modules) do
		if obj and obj.Disable then
			obj:Disable()
		end
	end
end

function CompactRunes:UpdateCDBars()
	if self.isDK and self.currentSpec and self.currentSpec ~= "" then
		local enabled = self.db.profile.cdbars[self.currentSpec].enabled
		local bar = self.cdbars[self.currentSpec]
		if enabled and not bar then
			bar = self.CooldownBar:Create(self.currentSpec)
			self.cdbars[self.currentSpec] = bar
		end
		for k, v in pairs(self.cdbars) do
			if k == self.currentSpec and enabled then
				v:Enable()
			elseif v and v.Disable then
				v:Disable()
			end
		end
	end
end

function addon.HasActiveTalent(talent)
	local activeGroup = _G.GetActiveSpecGroup()
	local id, name, iconTexture, selected, available = 
		_G.GetTalentInfoByID(addon.Talents[talent], activeGroup)
	return name and selected and available
end

function CompactRunes:SetSoulReaperPerc()
	addon.soulReaperPerc = 
		(_G.IsSpellKnown(SpellIds["Improved Soul Reaper"]) or 
			(addon.tierCount["T15 DPS"] or 0) >= 4) and 0.45 or 0.35
	if self.db.profile.debug then
		local srpFmt = "Soul Reaper Percent: %2.0f%%"
		self:Print(srpFmt:format((addon.soulReaperPerc * 100) or 0))
	end
end

function CompactRunes:UpdateAbilityProcsList()
	wipe(self.AbilityProcs)
	for spec, list in pairs(self.AbilityProcsList) do
		if spec == self.currentSpec or spec == "All" then
			for spell, flag in pairs(list) do
				local known, override = addon.IsSpellKnown(spell)
				if override then
					self.AbilityProcs[override] = flag
				else
					self.AbilityProcs[spell] = flag and known or false
				end
			end
		end
	end
end

function CompactRunes:UpdateRuneCDs()
	wipe(self.IgnoreRuneCDs)
	for spec, list in pairs(self.HasRuneCD) do
		if spec == self.currentSpec or spec == "All" then
			for spell, flag in pairs(list) do
				local known, override = addon.IsSpellKnown(spell)
				if known then
					local base = _G.GetSpellBaseCooldown(SpellIds[spell])
					if base and base > 0 then
						local grace = ceil(base / 1000 / 3)
						if grace > 2 then
							self.IgnoreRuneCDs[spell] = grace
						end
					end
				end
			end
		end
	end
end

function CompactRunes:CheckClass()
	self.isDK = false
  local class, className = _G.UnitClass("player")
  if className and (className == 'DEATH KNIGHT' or className == 'DEATHKNIGHT') then
		self.isDK = true
	end
end

function CompactRunes:CheckTalents()
	self:CheckClass()

	if self.isDK then
		local activeSpecNum = _G.GetSpecialization()
		if activeSpecNum and activeSpecNum > 0 then
			local id, name, desc, texture = _G.GetSpecializationInfo(activeSpecNum)
    	if texture == "Interface\\Icons\\Spell_Deathknight_BloodPresence" then
				self.currentSpec = "Blood"
			elseif texture == "Interface\\Icons\\Spell_Deathknight_FrostPresence" then
				self.currentSpec = "Frost"
			elseif texture == "Interface\\Icons\\Spell_Deathknight_UnholyPresence" then
				self.currentSpec = "Unholy"
			else
				self:Print("Error detecing player spec.")
				self.currentSpec = "Blood"
			end
		end

		if self.currentSpec ~= "Blood" and frames.boneBar then
			frames.boneBar.bar:Hide()
			frames.boneBar.bar.active = false
			frames.boneBar.bar:SetScript("OnUpdate", nil)
			frames.boneBar.bar.state = nil
		end

		local soulReaperId = SpellIds["Soul Reaper (Unholy)"]
		self.soulReaperKnown = soulReaperId and _G.IsSpellKnown(soulReaperId) or false
		self:SetSoulReaperPerc()

		self.outbreak = _G.IsSpellKnown(SpellIds["Outbreak"])

		self.regenMethod = "None"
		self.plagueLeech = false
		self.necroticPlague = false
		local activeSpecGroup = _G.GetActiveSpecGroup()

		self.hasBloodTap = false

		if activeSpecGroup then
			if addon.HasActiveTalent("Blood Tap") then
				self.regenMethod = "BT"
				self.hasBloodTap = true
			elseif addon.HasActiveTalent("Runic Empowerment") then
				self.regenMethod = "RE"
			elseif addon.HasActiveTalent("Runic Corruption") then
				self.regenMethod = "RC"
			end
			self.plagueLeech = addon.HasActiveTalent("Plague Leech")
			self.necroticPlague = addon.HasActiveTalent("Necrotic Plague")
		end
 	else
		self.regenMethod = "None"
	end

	self:CheckSpells()
	self:UpdateDTBCBars()
	for spell, enabled in pairs(addon.DynamicSpells) do
		local name, _, icon = _G.GetSpellInfo(spell)
		if icon then SpellIcons[spell] = icon end
	end
	self:UpdateCDBars()
	self:UpdateAbilityProcsList()
	self:UpdateRuneCDs()
	self:CheckForGhoul()
	for k, v in pairs(self.currentProcs) do
		if k and v and not addon.IsSpellKnown(k) then
			self.currentProcs[k] = nil
		end
	end
	if frames.powerBar then
		frames.powerBar:UpdateDisplay()
	end

	addon:FireCallback("TalentUpdate")

	if self.db.profile.debug then
		local fmt = "CheckTalents [isDK=%s,spec=%s,regen=%s,SR=%s,PL=%s,NP=%s]"
		self:Print(
			fmt:format(tostring(self.isDK),
				self.currentSpec, self.regenMethod, 
				tostring(self.soulReaperKnown),
				tostring(self.plagueLeech),
				tostring(self.necroticPlague)))
	end
end

function CompactRunes:CheckSpells()
	self.enabled.BoneShieldBar = self.db.profile.bars.BoneShieldBar.enabled and 
			self.currentSpec == "Blood" and _G.IsSpellKnown(SpellIds["Bone Shield"])
end

local function onUpdateRune(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.showing.timer = self.showing.timer + elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		self.lastUpdate = 0
		local profile = CompactRunes.db.profile
		self:SetValue(self.showing.timer)

		local remain = abs(self.showing.duration - self.showing.timer)
		local almostReadyTime = profile.almostReadyTime
		if almostReadyTime > 0 and almostReadyTime >= remain then
			self:SetAlpha(profile.almostReadyAlpha)
		end
		if profile.rechargeTimers ~= "None" then
			if self.showing.start > 0 and 
				(self.status ~= "depleted" or profile.rechargeTimers == "All") then
				self.value:SetText(tostring(round(remain)))
			else
				self.value:SetText("")
			end
		end
	end
end

local function onUpdateDisease(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.timer = self.timer - elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		self.lastUpdate = 0
		if self.active then
			if self.timer < 0 then
				self.timer = 0
				self.active = false
				--CompactRunes:CheckPlagueLeech()
			end
		
			self:Show()
			self:SetValue(self.timer)
			self.value:SetText(tostring(round(self.timer)))
		else
			self.value:SetText("0")
			self:Hide()
			self:SetScript("OnUpdate", nil)
			--CompactRunes:CheckPlagueLeech()
		end
	end
end

local function onUpdateNecrotic(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.timer = self.timer - elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		self.lastUpdate = 0
		local db = self.parent.db
		if self.active then
			if self.timer < 0 then
				self.timer = 0
				self.active = false
			end
		
			self:Show()
			if db.progress == "Time" then
				self:SetValue(self.timer)
			end
			if db.primary == "Time" then
				self.value:SetText(tostring(round(self.timer)))
			else
				self.secondary:SetText(tostring(round(self.timer)))
			end
		else
			if db.primary == "Time" then
				self.value:SetText("0")
			else
				self.secondary:SetText("0")
			end
			self:Hide()
			self:SetScript("OnUpdate", nil)
			--CompactRunes:CheckPlagueLeech()
		end
	end
end

local function onUpdateDarkTrans(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.timer = self.timer - elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		self.lastUpdate = 0
		if self.active then
			if self.timer < 0 then
				self.timer = 0
				self.active = false
				self:SetScript("OnUpdate", nil)
				self:Hide()
			else
				self:Show()
				self:SetValue(self.timer)
				if self.dt then
					self.value:SetText(tostring(round(self.timer)))
				end
			end
		else
			self:Hide()
		end
	end
end

local function onUpdateBloodCharge(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.timer = self.timer - elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		self.lastUpdate = 0
		if self.active then
			if self.timer < 0 then
				self.timer = 0
				self.active = false
				self:SetScript("OnUpdate", nil)
				self:Hide()
			else
				self:Show()
				self:SetValue(self.timer)
			end
		else
			self:Hide()
		end
	end
end

local function onUpdateBoneShield(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.timer = self.timer - elapsed
	self.recharge = self.recharge - elapsed
	if self.lastUpdate >= 0.1 then
		if self.active then
			local profile = CompactRunes.db.profile.bars["BoneShieldBar"]
			if self.timer > 0 then
				self:Show()
				if profile.progress == "Time" then
					self:SetValue(self.timer)
				elseif profile.progress == "Charges" then
					self:SetValue(self.count)
				end
			else
				self.timer = 0
				self.active = false
				self:SetScript("OnUpdate", nil)
				self:Hide()
			end
		elseif self.recharging then
			if self.recharge >= 0 then
				local remaining = 0
				if self.recharge > 60 then
					remaining = tostring(ceil(self.recharge / 60)) .. "m"
				else
					remaining = tostring(round(self.recharge))
				end
				self.rechargeTime:SetText(remaining)
			else
				self.recharging = false
				self.recharge = 0
				self.value:SetText("-")
				self:SetScript("OnUpdate", nil)	
			end
		else
			self:Hide()
		end
		self.lastUpdate = 0
	end
end

local function onUpdateInd(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		local outOfRange = false
		if _G.UnitExists("target") and not _G.UnitIsDead("target") and
			_G.UnitCanAttack("player", "target") then
			if _G.IsSpellInRange(SpellNames["Plague Strike"], "target") == 0 then
				outOfRange = true
			end
		end

		if outOfRange ~= self.outOfRange then
			self.outOfRange = outOfRange
			self:UpdateInd()
		end

		self.lastUpdate = 0
	end
end

local srTimer = 0
local function onUpdateIconTimer(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.timer = self.timer - elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		self.lastUpdate = 0
		if self.active then
			if self.timer < 0 then
				self.timer = 0
				self.active = false
				self:SetScript("OnUpdate", nil)
				self:SetDesiredAlpha()
			else
				--self:SetValue(self.timer)
				self.value:SetText(tostring(round(self.timer)))
			end
		else
			self:SetDesiredAlpha()
		end
		srTimer = self.timer
	end
end

local fontFlags = {}
function CompactRunes:GetFontSettings()
	wipe(fontFlags)
	for k, v in pairs(self.db.profile.fontFlags) do
		if v then tinsert(fontFlags, k) end
	end
	return self.db.profile.font_face, self.db.profile.font_size, tconcat(fontFlags, ",")
end

-- Define a generic class for the bars
Bar.__index = Bar
addon.Bar = Bar

function Bar:Create(name, fontAdj, width, height, setPoint, setValuePoint)
    local object = _G.setmetatable({}, Bar)
	object.name = name
	object.fontAdj = fontAdj
	
	local widthFunc = width
	if _G.type(width) ~= "function" then
		widthFunc = function() return width end
	end
	object.GetWidth = widthFunc
	local heightFunc = width
	if _G.type(height) ~= "function" then
		heightFunc = function() return height end
	end
	object.GetHeight = height

	object.SetPoint = setPoint
	object.SetDesiredPoint = setPoint
	object.SetValuePoint = setValuePoint
	object:Initialize()
	-- Add the bar to the addon's table of bars
	CompactRunes.bars[name] = object
	return object
end

function Bar:Initialize()
	local bar = _G.CreateFrame("StatusBar", 
		"CompactRunes_"  .. self.name, frames.main)
	self.bar = bar
	bar.parent = self
	self.db = CompactRunes.db.profile.bars[self.name]
	self.altcolor = false
	self.isRuneBar = self.name:find("RuneBar%d")
	bar:SetScale(1)
	if self.name == "BoneShieldBar" or 
		(self.isRuneBar and CompactRunes.db.profile.runeOrientation == "VERTICAL") then
		bar:SetOrientation("VERTICAL")
	else
		bar:SetOrientation("HORIZONTAL")
	end
	bar:SetWidth(self:GetWidth())
	bar:SetHeight(self:GetHeight())
	local bt = LSM:Fetch("statusbar", CompactRunes.db.profile.texture)
	bar:SetStatusBarTexture(bt)
	bar:GetStatusBarTexture():SetHorizTile(false)
	bar:GetStatusBarTexture():SetVertTile(false)
	if not self.isRuneBar then
		local bc = self.db.color
		bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
	end
	bar.bg = bar:CreateTexture(nil, "BACKGROUND")
	bar.bg:SetTexture(bt)
	bar.bg:SetAllPoints(true)
	if not self.isRuneBar then
		local bgc = self.db.bgcolor
		bar.bg:SetVertexColor(bgc.r, bgc.g, bgc.b, bgc.a)
	end

	local ff, fh, fflags = CompactRunes:GetFontSettings()
	local font = LSM:Fetch("font", ff)
	bar.value = bar:CreateFontString(nil, "OVERLAY")
	bar.value:SetPoint("CENTER")
	bar.value:SetFont(font, fh + self.fontAdj, fflags)
	bar.value:SetJustifyH("CENTER")
	--bar.value:SetShadowOffset(1, -1)
	local tc = self.db.textcolor
	bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	if self.isRuneBar and CompactRunes.db.profile.rechargeTimers == "None" then
		bar.value:Hide()
	end
	if self.name == "BloodShieldBar" then
		bar.value:SetText("0")
	end

	if self.name == "BoneShieldBar" then
		bar.rechargeTime = bar:CreateFontString(nil, "OVERLAY")
		bar.rechargeTime:SetPoint("CENTER")
		bar.rechargeTime:SetFont(font, fh + self.fontAdj - 1, fflags)
		bar.rechargeTime:SetJustifyH("CENTER")
		local tc = self.db.rechargeTextColor
		bar.rechargeTime:SetTextColor(tc.r, tc.g, tc.b, tc.a)
		bar.rechargeTime:SetText("0")
		bar.rechargeTime:Hide()
	end

	if self.name == "NecroticPlagueBar" then
		bar.secondary = bar:CreateFontString(nil, "OVERLAY")
		bar.secondary:SetFont(font, fh + self.fontAdj - 2, fflags)
		local tc = self.db.secondaryColor
		bar.secondary:SetTextColor(tc.r, tc.g, tc.b, tc.a)
		bar.secondary:SetText("0")

		self.UpdateCustom = function(self)
			local bar = self.bar
			local pos = (self.db.secondaryPos ~= "NONE" and self.db.secondaryPos)	
				or "RIGHT"
			bar.secondary:SetPoint(pos)
			bar.secondary:SetJustifyH(pos)
			if self.db.secondaryPos ~= "NONE" then
				bar.secondary:Show()
			else
				bar.secondary:Hide()			
			end
			if self.db.progress == "Stacks" then
				bar:SetMinMaxValues(0, addon.MAX_NP_STACKS)
			elseif self.db.progress == "None" then
				bar:SetMinMaxValues(0, 1)
				bar:SetValue(1)
			end
			if self.db.primary == "None" then
				bar.value:SetText("")
			end
		end
		self:UpdateCustom()
	end

	if self.SetPoint then self:SetPoint() end
	if self.SetValuePoint then
		self:SetValuePoint()
	else
	    bar.value:SetPoint("CENTER")
	end

	if self.name == "FrostFeverBar" or 
		self.name == "BloodPlagueBar" or 
		self.name == "DarkTransformationBar" or
		self.name == "BloodChargeBar" or
		self.name == "BloodShieldBar" or
		self.name == "BoneShieldBar" or
		self.name == "NecroticPlagueBar" or
		self.name == "SwingTimerBar" then
		bar.active = false
		bar.timer = 0
		bar:Hide()
	end

	if self.name == "BoneShieldBar" then
		self.bar.recharge = 0
	end

	bar.locked = true
end

function Bar:Show()
	self.bar:Show()
end

function Bar:Hide()
	self.bar:Hide()
end

function Bar:IsLocked()
	return self.bar.locked
end

function Bar:Lock(locked)
	if locked ~= nil then
		self.bar.locked = locked
	end
	self.bar:EnableMouse(not self.bar.locked)
	if self.name == "BloodShieldBar" then
		if not self.bar.locked then
			self.bar:Show()
		elseif not self.bar.active then
			self.bar:Hide()
		end
	end
end

function Bar:SetMovable()
	local bar = self.bar
    bar:SetMovable(true)
    bar:RegisterForDrag("LeftButton")
    bar:SetScript("OnDragStart",
        function(self, button)
			if not self.locked then
            	self:StartMoving()
			end
        end)
    bar:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			self.parent.db.x, self.parent.db.y = x, y
			self:SetUserPlaced(false);
        end)
	self:Lock()
end

local numberFmt = "%.0f"
function Bar:SetValue(value)
	self.bar:SetValue(value)
	self.bar.value:SetText(numberFmt:format(value))
end

function Bar:SetValueTextColor(color)
   	local tc = color or self.db.textcolor
   	self.bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
end

function Bar:Reset()
	self:ResetFonts()
	self:UpdateVisibility()
	--self:UpdateTexture()
	self:UpdateGraphics()
end

function Bar:ResetFonts()
	local ff, fh, fontFlags = CompactRunes:GetFontSettings()
	local font = LSM:Fetch("font", ff)
	self.bar.value:SetFont(font, fh + self.fontAdj, fontFlags)
	self.bar.value:SetText(self.bar.value:GetText())
	if self.name == "BoneShieldBar" then
		self.bar.rechargeTime:SetFont(font, fh + self.fontAdj - 1, fontFlags)
		self.bar.rechargeTime:SetText(self.bar.rechargeTime:GetText())
	elseif self.name == "NecroticPlagueBar" then
		self.bar.secondary:SetFont(font, fh + self.fontAdj - 1, fontFlags)
		self.bar.secondary:SetText(self.bar.secondary:GetText())
	end
end

function Bar:UpdateVisibility()
	if self.GetWidth then self.bar:SetWidth(self:GetWidth()) end
	if self.GetHeight then self.bar:SetHeight(self:GetHeight())	end
	if self.SetPoint and not self.isRuneBar then self:SetPoint() end
	if self.SetValuePoint then
		self:SetValuePoint()
	else
	    self.bar.value:SetPoint("CENTER")
	end
	if self.name == "BoneShieldBar" then
		if not (self.currentSpec == "Blood" or self.db.enabled) then
			self.bar:Hide()
			self.bar:SetScript("OnUpdate", nil)
		end
	else
		if not self.db.enabled and not self.isRuneBar then
			self.bar:Hide()
			self.bar:SetScript("OnUpdate", nil)
		end
	end

	self:UpdateTexture()
end

function Bar:UpdateTexture()
    local bt = LSM:Fetch("statusbar", CompactRunes.db.profile.texture)
	self.bar:SetStatusBarTexture(bt)
	self.bar.bg:SetTexture(bt)
    self.bar:GetStatusBarTexture():SetHorizTile(false)
    self.bar:GetStatusBarTexture():SetVertTile(false)
	--self:UpdateGraphics()
end

function Bar:UpdateGraphics()
	if not self.isRuneBar then
		local bc, bgc, tc
		if self.altcolor then
			bc = self.db.alt_color or self.db.color
			bgc = self.db.alt_bgcolor or self.db.bgcolor
			tc = self.db.alt_textcolor or self.db.textcolor
		else
			bc = self.db.color
			bgc = self.db.bgcolor
			tc = self.db.textcolor
		end

		self.bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
		self.bar.bg:SetVertexColor(bgc.r, bgc.g, bgc.b, bgc.a)
		self.bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
		if self.name == "BoneShieldBar" then
			local tc = self.db.rechargeTextColor
			self.bar.rechargeTime:SetTextColor(tc.r, tc.g, tc.b, tc.a)
		elseif self.name == "NecroticPlagueBar" then
			local tc = self.db.secondaryColor
			self.bar.secondary:SetTextColor(tc.r, tc.g, tc.b, tc.a)
		end
	end
end

local function GetRuneWidth()
	local profile = CompactRunes.db.profile
	local dimensions = profile.dimensions[profile.runeOrientation]
	return dimensions.width
end

local function GetRuneHeight()
	local profile = CompactRunes.db.profile
	local dimensions = profile.dimensions[profile.runeOrientation]
	return dimensions.height
end

local function GetRuneSpacing()
	local profile = CompactRunes.db.profile
	local dimensions = profile.dimensions[profile.runeOrientation]
	return dimensions.spacing
end

local function GetRuneInset()
	local profile = CompactRunes.db.profile
	local dimensions = profile.dimensions[profile.runeOrientation]
	return dimensions.inset
end

local function GetBarSpacing()
	local profile = CompactRunes.db.profile
	local dimensions = profile.dimensions[profile.runeOrientation]
	return dimensions.barSpacing
end

local function GetHalfBarHeight()
	local profile = CompactRunes.db.profile
	if profile.runeOrientation == "VERTICAL" then
		return GetRuneWidth()
	else
		return GetRuneHeight()
	end
end

local function GetHalfBarWidth()
	local profile = CompactRunes.db.profile
	if profile.runeOrientation == "VERTICAL" then
		return (GetRuneWidth() * 3 + GetRuneSpacing()) / 2 
	else
		return GetRuneWidth()
	end
end

local function GetBarHeight()
	local profile = CompactRunes.db.profile
	if profile.runeOrientation == "VERTICAL" then
		return GetRuneWidth()
	else
		return GetRuneHeight()
	end
end
addon.GetBarHeight = GetBarHeight

local function GetBarWidth()
	local profile = CompactRunes.db.profile
	if profile.runeOrientation == "VERTICAL" then
		return GetRuneWidth() * 3 + GetRuneSpacing() * 2
	else
		return GetRuneWidth() * 2 + GetRuneSpacing()
	end
end
addon.GetBarWidth = GetBarWidth

addon.depletedRunes = 0
addon.fullyDepletedRunes = 0
local runes = {
	[1] = {
		partner = 2,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[2] = {
		partner = 1,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[3] = {
		partner = 4,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[4] = {
		partner = 3,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[5] = {
		partner = 6,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[6] = {
		partner = 5,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
}
local runeBars = {}
local orderings = {
	["BUF"] = { 1, 2, 3, 4, 5, 6 },
	["BFU"] = { 1, 2, 5, 6, 3, 4 },
	["UFB"] = { 3, 4, 5, 6, 1, 2 },
	["UBF"] = { 3, 4, 1, 2, 5, 6 },
	["FBU"] = { 5, 6, 1, 2, 3, 4 },
	["FUB"] = { 5, 6, 3, 4, 1, 2 },
}
function CompactRunes:SetRuneOrder(ordering)
	for i = 1, 6 do
		local barNum = orderings[ordering][i]
		local runeBar = runeBars[barNum]
		if runeBar then
			local bar = runeBar.bar
			bar:ClearAllPoints()
			local inset = GetRuneInset()
			if self.db.profile.runeOrientation == "VERTICAL" then
				bar:SetPoint("BOTTOMLEFT", frames.main, "BOTTOMLEFT", 
					3 + -1 * (1 - inset - (GetRuneWidth() + 
					 GetRuneSpacing()) * (ceil(i / 2) - 1)),
					2 + inset + (GetRuneHeight() +
					 GetRuneSpacing()) * ((i - 1) % 2))
			else
				bar:SetPoint("TOPLEFT", frames.main, "TOPLEFT", 
					2 + inset + (GetRuneWidth() +
					 GetRuneSpacing()) * ((i - 1) % 2), 
					- 2 - inset - (GetRuneHeight() + 
					 GetRuneSpacing()) * (ceil(i / 2) - 1))
			end
		end
	end
end

function CompactRunes:ShowRechargeNumbers(value)
	for k, v in pairs(self.bars) do
		if v and v.isRuneBar and v.bar and v.bar.value then
			if value ~= "None" then
				v.bar.value:Show()
			else
				v.bar.value:Hide()
			end
		end
	end

	--for i = 1, 6 do
	--	local runeBar = runeBars[i]
	--	if runeBar then
	--		local bar = runeBar.bar
	--		bar.value:Show()
	--	end
	--end
end

function CompactRunes:UpdateTargetMaxHealth()
	self.targetMaxHealth = _G.UnitHealthMax("target") or 0
	self:UpdateTargetHealth()
end

function CompactRunes:UpdateTargetHealth()
	self.targeHealth = UnitHealth("target") or 0
	self.targetPercent = self.targetMaxHealth > 0 and 
		self.targeHealth / self.targetMaxHealth or 0
	self:CheckSoulReaper()
end

local lastSRUpdate = 0
function CompactRunes:CheckSoulReaper(noUpdate, srUsed)
	local update = false
	if not self.soulReaperKnown then return false end

	local start, duration, enabled = GetSpellCooldown(self.soulReaperIds[self.currentSpec])
	lastSRUpdate = start

	if srUsed then
		local now = GetTime()
		--local fmt = "SR [%s, %s, %s]"
		--self:Print(fmt:format(tostring(start), tostring(duration), tostring(now - start)))

		local srCD = addon.SRCooldown
		if duration > addon.gcd and duration < addon.runeCD then
			srCD = duration - (now - start)
		end
		srTimer = srCD

		local ind = frames.srcdInd
		if ind and ind.enabled then
			ind.active = true
			ind.timer = srCD
			ind:SetScript("OnUpdate", onUpdateIconTimer)
			ind:SetDesiredAlpha()
		end
	end
	if duration == 0 then
		srTimer = 0
		local ind = frames.srcdInd
		if ind then
			ind.active = false
			ind.timer = 0
			ind:SetScript("OnUpdate", nil)
			ind:SetDesiredAlpha()
		end
	end	

	if self.db.profile.enabled[self.currentSpec]["Soul Reaper"] then
		local ignore = false
		local srActive = false
		if self.srMethod == "cooldown" then
			-- local start, duration, enabled =
			-- 	GetSpellCooldown(self.soulReaperIds[self.currentSpec])

			srActive = srTimer > 0
			-- if duration >= 10 and self.db.profile.ignoreRuneCDs["Soul Reaper"] then
			-- 	ignore = true
			-- elseif duration > 1 then
			-- 	srActive = true
			-- end
		else
			srActive = self.targetSoulReaper
		end

		if not ignore then
			local old = self.currentProcs["Soul Reaper"]
			if self.targetHostile and self.targetPercent < addon.soulReaperPerc and 
				not srActive then
				self.currentProcs["Soul Reaper"] = 1
			else
				self.currentProcs["Soul Reaper"] = nil
			end
			if old ~= self.currentProcs["Soul Reaper"] then
				if not noUpdate then self:UpdateProcIcon() end
				update = true
			end
		end
	end
	return update
end

function CompactRunes:UpdateRunePair(runeIndex)
	local profile = self.db.profile
	local partnerIndex = runes[runeIndex].partner

	local time = GetTime()
	local firstIndex = min(runeIndex, partnerIndex)
	local secondIndex = max(runeIndex, partnerIndex)
	local first = runeBars[firstIndex]
	local second = runeBars[secondIndex]
	if profile.dimensions[profile.runeOrientation].swap then
		local tmp = first
		first = second
		second = tmp
	end
	local sooner, later = runes[firstIndex], runes[secondIndex]
	if time - sooner.start < time - later.start then
		sooner, later = runes[secondIndex], runes[firstIndex]
	end

	local rc, rcc
	-- Update first
	first.bar.showing = sooner
	first.bar.status = sooner.ready and "ready" or "recharging"
	rc = RuneColors[RuneMapping[sooner.type]]
	if profile.runeBackground == 1 then
		rcc = RuneCooldownColors[RuneMapping[sooner.type]]
		rcc.a = profile.runeBackgroundAlpha or rcc.a
	else
		rcc = self.altRuneBkg
	end
	first.bar:SetStatusBarColor(rc.r, rc.g, rc.b, 1)
	first.bar.bg:SetVertexColor(rcc.r, rcc.g, rcc.b, rcc.a)
	first.bar:SetMinMaxValues(0, sooner.duration)
	first.bar:SetValue(sooner.timer)
	if sooner.ready == false then
		first.bar:SetScript("OnUpdate", onUpdateRune)
		first.bar:SetAlpha(profile.rechargeAlpha)
	else
		first.bar:SetScript("OnUpdate", nil)
		first.bar:SetAlpha(1.0)
		if profile.rechargeTimers ~= "None" then
			first.bar.value:SetText("")
		end
	end
	-- Update second
	second.bar.showing = later
	second.bar.status = sooner.ready and "recharging" or "depleted"
	rc = RuneColors[RuneMapping[later.type]]
	if profile.runeBackground == 1 then
		rcc = RuneCooldownColors[RuneMapping[later.type]]
		rcc.a = profile.runeBackgroundAlpha or rcc.a
	else
		rcc = self.altRuneBkg
	end
	second.bar:SetStatusBarColor(rc.r, rc.g, rc.b, 1)
	second.bar.bg:SetVertexColor(rcc.r, rcc.g, rcc.b, rcc.a)
	second.bar:SetMinMaxValues(0, later.duration)
	second.bar:SetValue(later.timer)
	if later.ready == false then
		second.bar:SetScript("OnUpdate", onUpdateRune)
		second.bar:SetAlpha(profile.rechargeAlpha)
	else
		second.bar:SetScript("OnUpdate", nil)
		second.bar:SetAlpha(1.0)
		if profile.rechargeTimers ~= "None" then
			second.bar.value:SetText("")
		end
	end
end

function CompactRunes:UpdateRuneType(runeIndex)
	local type = _G.GetRuneType(runeIndex)
	runes[runeIndex].type = type
	self:UpdateRunePair(runeIndex)
end

function CompactRunes:UpdateRuneCooldown(runeIndex, isEnergized)
	local profile = self.db.profile
	local start, duration, ready = _G.GetRuneCooldown(runeIndex)
	start = start or 0
	duration = duration or 0
	local type = _G.GetRuneType(runeIndex)
	local rune = runes[runeIndex]
	rune.ready = ready

	if rune.ready and self.regenMethod == "RE" then
		local timeLeft = rune.timer - rune.duration
		if timeLeft > 0.5 then
			self.TotalStats.reProcs[type] = (self.TotalStats.reProcs[type] or 0) + 1
			self.LastFightStats.reProcs[type] = (self.LastFightStats.reProcs[type] or 0) + 1
			if profile.debug then
				self:Print("RE Proc - " .. (RuneMapping[type] or "Unknown"))
			end
		end
		--local refFmt = "Rune %d: %0.2f [%0.2f, %0.2f]"
		--self:Print(refFmt:format(runeIndex, timeLeft, 
		--	rune.duration, rune.timer))
	end

	local time = GetTime()
	rune.start = start
	if ready == false then
		rune.duration = duration
		rune.timer = time - start
	else
		rune.timer = 10
		rune.duration = 10
	end

	self:UpdateRunePair(runeIndex)

	--self:CheckSoulReaper()

	-- Updated count of depleted runes
	local depleted = 0
	local fullyDepleted = 0
	for i, v in ipairs(runes) do
		if v then
			if not v.ready then
				depleted = depleted + 1
				if v.partner < i then
					local partner = runes[v.partner]
					if partner and not partner.ready then
						fullyDepleted = fullyDepleted + 1
					end
				end
			end
		end
	end
	addon.depletedRunes = depleted
	addon.fullyDepletedRunes = fullyDepleted
	--local fmt = "Depleted: %s / %s"
	--self:Print(fmt:format(tostring(depleted), tostring(fullyDepleted)))
	self:CheckPlagueLeech()
end

function CompactRunes:UpdateRunes()
	for i = 1, 6 do
		self:UpdateRuneType(i)
		self:UpdateRuneCooldown(i)
	end	
end

function CompactRunes:CreateSRCDIndicator()
	local bar = _G.CreateFrame("Frame", "CompactRunes_SRCD", frames.main)
	bar.db = self.db.profile
	frames.srcdInd = bar
	bar.active = false
	bar.enabled = CompactRunes.db.profile.srcdIcon.enabled
	bar.SetDesiredAlpha = function(bar)
		local db = CompactRunes.db.profile.srcdIcon
		bar.enabled = db.enabled
		if bar.enabled then
			local profile = db.alphas
			bar:SetAlpha(bar.active and profile.active or 
				(_G.UnitAffectingCombat("player") and profile.combat or profile.ooc))
			if bar.active then bar.value:Show() else bar.value:Hide() end
		else
			bar:Hide()
		end
	end
	bar.SetDesiredPoint = function(bar)
		bar:ClearAllPoints()
		local profile = CompactRunes.db.profile
		bar:SetPoint("RIGHT", frames.procInd, "LEFT", -5, 0)
		-- local x, y = profile.procIcon.x, profile.procIcon.y
		-- if profile.procIcon.userPlaced and x ~= nil and y ~= nil then
		-- 	    	bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
		-- else
		-- 	    	bar:SetPoint("BOTTOM", frames.main, "TOP", 0,
		-- 		GetRuneSpacing() * 7 + GetBarHeight() + 2)
		-- end
	end
	bar:SetDesiredPoint()
	bar:SetScale(1)
	bar.GetDesiredHeight = function()
		local scale = self.db.profile.procIcon.scale
		return ceil(scale*(frames.main:GetWidth()*frames.main:GetHeight()*0.233)^0.5)
	end
	bar.GetDesiredWidth = function()
		local scale = self.db.profile.procIcon.scale
		return ceil(scale*(frames.main:GetWidth()*frames.main:GetHeight()*0.233)^0.5)
	end
	bar.SetDimensions = function(bar)
	    bar:SetWidth(bar:GetDesiredHeight())
	    bar:SetHeight(bar:GetDesiredWidth())
	end
	bar:SetDimensions()
	bar.texture = bar:CreateTexture(nil, "ARTWORK")
	bar.texture:SetAllPoints(bar)
	bar.texture:SetVertexColor(1, 1, 1)
	bar.texture:SetTexture(SpellIcons["Soul Reaper"])
	local ff, fh, fflags = self:GetFontSettings()
	local font = LSM:Fetch("font", ff)
	bar.value = bar:CreateFontString(nil, "OVERLAY")
	bar.value:SetPoint("CENTER", bar, "CENTER")
	bar.value:SetFont(font, fh + 3, fflags)
	bar.value:SetJustifyH("CENTER")
	bar.value:SetJustifyV("CENTER")
	bar.value:SetShadowOffset(1, -1)
	local tc = bar.db.bars.ProcInd.textcolor
	bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	bar.value:Show()
	
	bar.ResetFonts = function()
		local ff, fh, fontFlags = self:GetFontSettings()
		local font = LSM:Fetch("font", ff)
		bar.value:SetFont(font, fh + 3, fontFlags)						
		bar.value:SetText(bar.value:GetText())
	end

	bar.locked = true
	bar.Lock = function(self, locked)
		if locked ~= nil then
			self.locked = locked
		end
		self:EnableMouse(not self.locked)
		self:Configure(not self.locked)
	end
	bar.IsLocked = function(self)
		return self.locked
	end

	bar.configMode = false
	bar.Configure = function(self, config)
		if config ~= nil then
			self.configMode = config
		else
			self.configMode = not self.configMode
		end
		if self.configMode then
			self:Show()
		else
			--CompactRunes:UpdateProcIcon()
		end
	end

    bar:SetMovable(true)
    bar:RegisterForDrag("LeftButton")
    bar:SetScript("OnDragStart",
			function(self, button)
				if not self.locked then
					self:StartMoving()
				end
			end)
    bar:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			local procIcon = CompactRunes.db.profile.procIcon
			procIcon.x, procIcon.y = x, y
			procIcon.userPlaced = true
			self:SetUserPlaced(false)
        end)
	bar:Lock()
	bar:SetDesiredAlpha()
	bar:Show()
end

function CompactRunes:CreateProcIndicator()
	local bar = _G.CreateFrame("Frame", "CompactRunes_Proc", frames.main)
	bar.db = self.db.profile
	frames.procInd = bar
	bar.currentProc = nil
	bar.procList = nil
	bar.SetDesiredAlpha = function(bar)
		local profile = CompactRunes.db.profile.procIcon.alphas
		bar:SetAlpha(_G.UnitAffectingCombat("player") and profile.combat or profile.ooc)
	end
	bar.SetDesiredPoint = function(bar)
		bar:ClearAllPoints()
		local profile = CompactRunes.db.profile
		local x, y = profile.procIcon.x, profile.procIcon.y
		if profile.procIcon.userPlaced and x ~= nil and y ~= nil then
	    	bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
		else
	    	bar:SetPoint("BOTTOM", frames.main, "TOP", 0, 
				GetRuneSpacing() * 7 + GetBarHeight() + 2)
		end
	end
	bar:SetDesiredPoint()
	bar:SetScale(1)
	bar.GetDesiredHeight = function()
		local scale = self.db.profile.procIcon.scale
		return ceil(scale*(frames.main:GetWidth()*frames.main:GetHeight()*0.233)^0.5)
	end
	bar.GetDesiredWidth = function()
		local scale = self.db.profile.procIcon.scale
		return ceil(scale*(frames.main:GetWidth()*frames.main:GetHeight()*0.233)^0.5)
	end
	bar.SetDimensions = function(bar)
	    bar:SetWidth(bar:GetDesiredHeight())
	    bar:SetHeight(bar:GetDesiredWidth())
	end
	bar:SetDimensions()
	bar.texture = bar:CreateTexture(nil, "ARTWORK")
    bar.texture:SetAllPoints(bar)
    bar.texture:SetVertexColor(1, 1, 1)
	local ff, fh, fflags = self:GetFontSettings()
	local font = LSM:Fetch("font", ff)
    bar.value = bar:CreateFontString(nil, "OVERLAY")
    bar.value:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", 0, -2)
    bar.value:SetFont(font, fh + 3, fflags)
    bar.value:SetJustifyH("CENTER")
	bar.value:SetJustifyV("CENTER")
    bar.value:SetShadowOffset(1, -1)
    local tc = bar.db.bars.ProcInd.textcolor
    bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	bar.value:Hide()
	
	bar.ResetFonts = function()
		local ff, fh, fontFlags = self:GetFontSettings()
		local font = LSM:Fetch("font", ff)
		bar.value:SetFont(font, fh + 3, fontFlags)						
		bar.value:SetText(bar.value:GetText())
	end

	bar.locked = true
	bar.Lock = function(self, locked)
		if locked ~= nil then
			self.locked = locked
		end
		self:EnableMouse(not self.locked)
		self:Configure(not self.locked)
	end
	bar.IsLocked = function(self)
		return self.locked
	end

	bar.configMode = false
	bar.Configure = function(self, config)
		if config ~= nil then
			self.configMode = config
		else
			self.configMode = not self.configMode
		end
		if self.configMode then
			self.texture:SetTexture(SpellIcons["Auto Attack"])
			self:Show()
		else
			CompactRunes:UpdateProcIcon()
		end
	end

    bar:SetMovable(true)
    bar:RegisterForDrag("LeftButton")
    bar:SetScript("OnDragStart",
        function(self, button)
			if not self.locked then
            	self:StartMoving()
			end
        end)
    bar:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			local procIcon = CompactRunes.db.profile.procIcon
			procIcon.x, procIcon.y = x, y
			procIcon.userPlaced = true
			self:SetUserPlaced(false)
        end)
	bar:Lock()
	self:UpdateProcIcon()
end

function CompactRunes:CheckPlagueLeech(trigger)
	-- TODO:
	-- Verify the criteria of PL:
	-- Plague Leech is off cooldown
	-- 2 depleted runes, but optional
	-- NP or both BP and FF
	-- Outbreak off CD if user requires it
	local activate = nil

	if trigger == nil then
		if self.plagueLeech then
			local start, duration, enabled = GetSpellCooldown(SpellNames["Plague Leech"])
			if duration and duration <= 1 then
				local diseased = (self.necroticPlague and addon.diseaseCount >= 1) or 
						addon.diseaseCount > 1
				if diseased and (not self.db.profile.plRequireRunes or 
					addon.fullyDepletedRunes >= 2) then
					if self.outbreak and self.db.profile.plOutbreak then
						local s, d, e = GetSpellCooldown(SpellNames["Outbreak"])
						if d and d <= 1 then activate = true end
					else
						activate = true
					end
				end
			end
		end
	else
		activate = trigger or nil
	end

	activate = activate and 1 or nil
	if self.currentProcs["Plague Leech"] ~= activate then
		self.currentProcs["Plague Leech"] = activate
		self:UpdateProcIcon()
	end
end

function CompactRunes:UpdateProcIcon()
	local procInd = frames.procInd
	if not procInd then return end
	local found = false

	if self.db.profile.procs then
		local procPriority = self.db.profile.priorityList[self.currentSpec]
		local enabled = self.db.profile.enabled[self.currentSpec]

		if procPriority and enabled then
			for i, spell in ipairs(procPriority) do
				local s = spell
				local override = SpellOverridenByTalent[spell]
				if override and addon.HasActiveTalent(override) then s = override end
				local count = self.currentProcs[s]
				if enabled[spell] and count then
					found = true
					procInd.texture:SetTexture(SpellIcons[s])
					procInd:Show()
					if count > 1 then
						procInd.value:SetText(tostring(count))
						procInd.value:Show()
					else
						procInd.value:Hide()
					end

					break
				end
			end
		end
	end

	if not found then
		procInd:Hide()
	end
end

function CompactRunes:SetMainFrameBorder()
	local frame = frames.main
	if frame then
		local border = LSM:Fetch("border", self.db.profile.border)
		local backdrop = {
	        bgFile = "Interface\\Buttons\\WHITE8X8", 
	        edgeFile = border, tile = false, tileSize = 16, edgeSize = 16, 
	        insets = { left = 4, right = 4, top = 4, bottom = 4 }
	    }
		local bdbc = self.db.profile.frame.bdcolor
		local bdc = self.db.profile.frame.color
	    frame:SetBackdrop(backdrop)
	    frame:SetBackdropBorderColor(bdbc.r, bdbc.g, bdbc.b, bdbc.a)
	    frame:SetBackdropColor(bdc.r, bdc.g, bdc.b, bdc.a)

		border = LSM:Fetch("border", self.db.profile.borderInd)
		backdrop = {
	        bgFile = "Interface\\Buttons\\WHITE8X8", 
	        edgeFile = border, tile = false, tileSize = 16, edgeSize = 16, 
	        insets = { left = 4, right = 4, top = 4, bottom = 4 }
	    }
		local rc = self.db.profile.frame.rccolor
	    frame.ind:SetBackdrop(backdrop)
	    frame.ind:SetBackdropBorderColor(rc.r, rc.g, rc.b, rc.a)
	    frame.ind:SetBackdropColor(0, 0, 0, 0)
	end
end

function CompactRunes:IsBothDTBCBars()
	return self.currentSpec == "Unholy" and 
			self.regenMethod == "BT" and
			self.db.profile.bars.BloodChargeBar.enabled and
			self.db.profile.bars.DarkTransformationBar.enabled
end

function CompactRunes:UpdateDTBCBars()
	if self.isDK then
		local bars = {
			"DarkTransformationBar",
			"BloodChargeBar"
		}
		for k, v in pairs(bars) do
			if v and self.bars[v] then
				local bar = self.bars[v]
				bar:SetPoint()
				if not bar.db.enabled then
					bar.bar.active = false 
					bar.bar.timer = 0
					bar.bar:SetScript("OnUpdate", nil)
					bar.bar:Hide()
				end
			end
		end
	end
end

function CompactRunes:CalculateSize()
	self.sizes = self.sizes or {}
	self.sizes.Display = self.sizes.Display or {}
	self.sizes.Display.width = (GetRuneWidth() + GetRuneSpacing()) * 2 + 9
	self.sizes.Display.height = (GetRuneHeight() + GetRuneSpacing()) * 3 + 9
	self.sizes.hinset = 0
	self.sizes.vinset = 0
end

function CompactRunes:CreateValueBlock(name, parent, setPoint)
	if self.blocks[name] then return end
	local parentFrame = parent or _G.UIParent
	local frame = _G.CreateFrame("Frame", "CompactRunes_ValBlock_"..name, parentFrame)
	self.blocks[name] = frame
	frame.parentFrame = parentFrame
	frame.db = self.db.profile.blocks[name]
	frame.name = name
	frame.lock = true
	frame:SetBackdrop({
		bgFile = "Interface\\Buttons\\WHITE8X8", 
		edgeFile = "",
		tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 0, right = 0, top = 0, bottom = 0 }
		})

    frame.value = frame:CreateFontString(nil, "OVERLAY")
    frame.value:SetPoint("CENTER")
    frame.value:SetJustifyH("CENTER")
	frame.value:Show()

	if setPoint and _G.type(setPoint) == "function" then
		frame.SetDesiredPoint = setPoint
	else
		frame.SetDesiredPoint = function(self)
			self:SetPoint("CENTER")
		end
	end

	frame.SetValue = function(self, value) self.value:SetText(value) end

	frame.UpdateColor = function(self)
		local c
		if self.db.background then
			if not c then c = self.db.bgcolor end
			self:SetBackdropColor(c.r, c.g, c.b, c.a)
		else
			if not c then c = self.db.textcolor end
			self.value:SetTextColor(c.r, c.g, c.b, c.a)
		end
	end

	frame.ResetFonts = function(self)
		local ff, fh, fflags = CompactRunes:GetFontSettings()
		local font = LSM:Fetch("font", ff)
	    self.value:SetFont(font, self.db.fontSize or fh, fflags)
		self.value:SetText(self.value:GetText())
	end

	frame.IsLocked = function(self)
		return self.lock
	end

	frame.Lock = function(self, locked)
		if locked ~= nil then
			self.lock = locked
		end
		self:EnableMouse(not self.lock)
	end

	frame.UpdateLayout = function(self)
		self:SetHeight(self.db.height)
		self:SetWidth(self.db.width)
		self:Lock()
		self:SetDesiredPoint()
		self:ResetFonts()
		if not self.db.background then
			self:SetBackdropColor(0,0,0,0)
		else
			local c = self.db.textcolor
			self.value:SetTextColor(c.r, c.g, c.b, c.a)
		end
		if self.db.enabled then self:Show() else self:Hide() end
	end

    frame:SetMovable(true)
    frame:RegisterForDrag("LeftButton")
    frame:SetScript("OnDragStart",
        function(self, button)
			if not self.lock then
            	self:StartMoving()
			end
        end)
    frame:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			self.db.x, self.db.y = x, y
			self:SetUserPlaced(false);
        end)

	frame:UpdateColor()
	frame:UpdateLayout()

	return frame
end

function CompactRunes:CreateBlocks()
	frames.ffStr = self:CreateValueBlock("FrostFeverStrength", 
		frames.ffBar.bar, 
		function(self)
			local db = CompactRunes.db.profile
			self:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				if db.diseaseOrder == "BF" then
					self:SetPoint("LEFT", self.parentFrame, "RIGHT", 
						self.db.offset, 0)
				else
					self:SetPoint("RIGHT", self.parentFrame, "LEFT", 
						-self.db.offset, 0)
				end
			end
		end)
	self.diseases.ff.val = frames.ffStr
	frames.bpStr = self:CreateValueBlock("BloodPlagueStrength", 
		frames.bpBar.bar, 
		function(self)
			local db = CompactRunes.db.profile
			self:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				if db.diseaseOrder == "BF" then
					self:SetPoint("RIGHT", self.parentFrame, "LEFT", 
						-self.db.offset, 0)
				else
					self:SetPoint("LEFT", self.parentFrame, "RIGHT", 
						self.db.offset, 0)
				end
			end
		end)
	self.diseases.bp.val = frames.bpStr
end

function CompactRunes:CreateDisplay()
	local frame = _G.CreateFrame("Frame", "CompactRunes_Display", _G.UIParent)
	frames.main = frame
	frame.db = self.db.profile.frame
	frame.SetDesiredPoint = function(self)
		self:SetPoint("CENTER", _G.UIParent, "CENTER", self.db.x, self.db.y)
	end
	frame:SetDesiredPoint()
	frame.GetDesiredWidth = function()
		local profile = self.db.profile
		if profile.runeOrientation == "VERTICAL" then
			return (GetRuneWidth() + GetRuneSpacing()) * 3 + GetRuneInset() * 2
		else
			return (GetRuneWidth() + GetRuneSpacing()) * 2 + GetRuneInset() * 2
		end
	end
	frame.GetDesiredHeight = function()
		local profile = self.db.profile
		if profile.runeOrientation == "VERTICAL" then
			return (GetRuneHeight() + GetRuneSpacing()) * 2 + GetRuneInset() * 2
		else
			return (GetRuneHeight() + GetRuneSpacing()) * 3 + GetRuneInset() * 2
		end
	end
	frame:SetWidth(frame.GetDesiredWidth())
	frame:SetHeight(frame.GetDesiredHeight())

	frame.ind = _G.CreateFrame("Frame", nil,frame)
	frame.ind:SetAllPoints(frame)
	frame.ind:Hide()
	frame.rcActive = false
	frame.outOfRange = false
	frame.UpdateInd = function(self)
		if self.outOfRange then
			local bc = CompactRunes.db.profile.rangeColor
			self.ind:SetBackdropBorderColor(bc.r, bc.g, bc.b, bc.a)
			self.ind:Show()
		elseif self.rcActive and CompactRunes.db.profile.rcProcInd then
			local bc = CompactRunes.db.profile.rcProcColor
			self.ind:SetBackdropBorderColor(bc.r, bc.g, bc.b, bc.a)
			self.ind:Show()
		else
			self.ind:Hide()
		end
	end
	frame.UpdateBackdropColor = function(self)
		local bdc = self.db.color
    self:SetBackdropColor(bdc.r, bdc.g, bdc.b, bdc.a)
	end	
	self:SetMainFrameBorder()

	frame.lock = self.db.profile.locked
    frame:SetMovable(true)
    frame:RegisterForDrag("LeftButton")
    frame:SetScript("OnDragStart",
        function(self, button)
			if not self.lock then
            	self:StartMoving()
			end
        end)
    frame:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			self.db.x, self.db.y = x, y
			self:SetUserPlaced(false);
        end)
    frame:EnableMouse(not frame.lock)

	frames.boneBar = Bar:Create("BoneShieldBar", 0, 
		function()
			local profile = self.db.profile
			if profile.runeOrientation == "VERTICAL" then
				return GetRuneWidth() - 2
			else
				return GetRuneHeight() - 2
			end
		end,
		function() return frame:GetDesiredHeight() end,
		function(self)
			self.bar:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self.bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				self.bar:SetPoint("BOTTOMLEFT", frames.main, "BOTTOMRIGHT", 
					GetBarSpacing() - 3, 0)
			end
		end
	)
	frames.boneBar:SetMovable()
	self:UpdateBoneShieldBarMode()

	frames.powerBar = Bar:Create("RunicPowerBar", 2, 
		function(self)
			return self.db.overrideWidth and self.db.width or GetBarWidth()
		end,
		function(self)
			return self.db.overrideHeight and self.db.height or GetBarHeight()
		end,
		function(self)
			self.bar:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self.bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				self.bar:SetPoint("TOP", frames.main, "BOTTOM", 
					0, -2)
			end
		end
	)
	frames.powerBar.UpdateDisplay = function(self)
		if self.db.showNumericValue then
			self.bar.value:Show()
		else
			self.bar.value:Hide()
		end
		self.tickmarks = self.tickmarks or {}
		local profile = CompactRunes.db.profile.rpTicks
		local spec = profile.specs[CompactRunes.currentSpec]
		if profile.enabled and spec then
			local cost = spec.cost
			local maxRP = _G.UnitPowerMax("player", 6)
			local numTicks = _G.math.floor(maxRP / cost)
			if (numTicks * cost) / maxRP > 0.9 then numTicks = numTicks - 1 end
			local interval = self:GetWidth() / (maxRP/cost)
			local c = profile.color
			for i = 1, numTicks do
				self.tickmarks[i] = self.tickmarks[i] or self.bar:CreateTexture()
				local tick = self.tickmarks[i]
				tick:SetTexture(c.r, c.g, c.b, c.a)
				tick:SetHeight(self:GetHeight())
				tick:SetWidth(profile.width)
				tick:SetPoint("LEFT", self.bar, "LEFT", interval * i, 0)
				tick:SetDrawLayer("OVERLAY")
				tick:Show()
			end
			for i = numTicks + 1, 5 do
				local tick = self.tickmarks[i]
				if tick then tick:Hide() end
			end
		else
			for i = 1, 5 do
				local tick = self.tickmarks[i]
				if tick then tick:Hide() end
			end
		end
	end
	frames.powerBar:SetMovable()
	self:UpdateRunicPowerMax()
	self:UpdateRunicPower()
	self:ShowRunicPowerBar()

	frames.bsBar = Bar:Create("BloodShieldBar", 2, 
		function(self)
			return self.db.overrideWidth and self.db.width or GetBarWidth()
		end,
		function(self)
			return self.db.overrideHeight and self.db.height or GetBarHeight()
		end,
		function(self)
			self.bar:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self.bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				self.bar:SetPoint("TOP", frames.main, "BOTTOM", 
					0, -2 - GetBarHeight() - 6)
			end
		end
	)
	frames.bsBar:SetMovable()

	frames.ffBar = Bar:Create("FrostFeverBar", 2, 
		function(self)
			return self.db.overrideWidth and self.db.width or GetHalfBarWidth()
		end,
		function(self)
			return self.db.overrideHeight and self.db.height or GetHalfBarHeight()
		end,
		function(self)
			local db = CompactRunes.db.profile
			self.bar:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self.bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				if db.diseaseOrder == "BF" then
					self.bar:SetPoint("BOTTOMRIGHT", frames.main, "TOPRIGHT", 
						-2 + -1 * GetRuneInset(), 0)
				else
					self.bar:SetPoint("BOTTOMLEFT", frames.main, "TOPLEFT", 
						2 + GetRuneInset(), 0)
				end
			end
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)
	frames.ffBar:SetMovable()
	self.diseases.ff.bar = frames.ffBar

	frames.bpBar = Bar:Create("BloodPlagueBar", 2, 
		function(self)
			return self.db.overrideWidth and self.db.width or GetHalfBarWidth()
		end,
		function(self)
			return self.db.overrideHeight and self.db.height or GetHalfBarHeight()
		end,
		function(self)
			self.bar:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self.bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				local db = CompactRunes.db.profile
				if db.diseaseOrder == "FB" then
					self.bar:SetPoint("BOTTOMRIGHT", frames.main, "TOPRIGHT", 
						-2 + -1 * GetRuneInset(), 0)
				else
					self.bar:SetPoint("BOTTOMLEFT", frames.main, "TOPLEFT", 
						2 + GetRuneInset(), 0)
				end
			end
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)
	frames.bpBar:SetMovable()
	self.diseases.bp.bar = frames.bpBar

	frames.npBar = Bar:Create("NecroticPlagueBar", 2, 
		function(self)
			return self.db.overrideWidth and self.db.width or GetBarWidth()
		end,
		function(self)
			return self.db.overrideHeight and self.db.height or GetBarHeight()
		end,
		function(self)
			local db = CompactRunes.db.profile
			self.bar:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self.bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				self.bar:SetPoint("BOTTOM", frames.main, "TOP", 
					0, 0)
			end
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)
	frames.npBar:SetMovable()

	frames.dtBar = Bar:Create("DarkTransformationBar", 2, 
		function(self)
			return self.db.overrideWidth and self.db.width or GetHalfBarWidth()
		end,
		function(self)
			return self.db.overrideHeight and self.db.height or GetHalfBarHeight()
		end,
		function(self)
			self.bar:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self.bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				local profile = CompactRunes.db.profile
				local yOffset = GetRuneSpacing() * 2 + GetBarHeight() + 2
				if CompactRunes:IsBothDTBCBars() then
					self.bar:SetPoint("BOTTOMLEFT", frames.main, "TOPLEFT", 
						GetRuneInset(), yOffset)
				else
					self.bar:SetPoint("BOTTOM", frames.main, "TOP", 
						0, yOffset)
				end
			end
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)
	frames.dtBar:SetMovable()

	frames.bcBar = Bar:Create("BloodChargeBar", 2, 
		function(self)
			return self.db.overrideWidth and self.db.width or GetHalfBarWidth()
		end,
		function(self)
			return self.db.overrideHeight and self.db.height or GetHalfBarHeight()
		end,
		function(self)
			self.bar:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self.bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				local profile = CompactRunes.db.profile
				local yOffset = GetRuneSpacing() * 2 + GetBarHeight() + 2
				if CompactRunes:IsBothDTBCBars() then
					self.bar:SetPoint("BOTTOMRIGHT", frames.main, "TOPRIGHT", 
						-2 + -1 * GetRuneInset(), yOffset)
				else
					self.bar:SetPoint("BOTTOM", frames.main, "TOP", 
						0, yOffset)
				end
			end
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)
	frames.bcBar:SetMovable()

	self:CreateProcIndicator()
	self:CreateSRCDIndicator()
	--self:CreateBlocks()

	-- Create the runes and then update their status
	for i = 1, 6 do
		runeBars[i] = Bar:Create("RuneBar"..i, 0, 
			function()
				return GetRuneWidth()
			end,
			function()
				return GetRuneHeight()
			end,
			nil,
			nil
		)
		runeBars[i].bar.showing = runes[i]
	end
	self:UpdateRunes()
	self:SetRuneOrder(self.db.profile.runeOrder)
	self:SetDisplayAlpha()
	frames.main:Show()
end

function CompactRunes:ShowRunicPowerBar()
	if self.db.profile.bars.RunicPowerBar.enabled then
		frames.powerBar.bar:Show()
		self:RegisterEvent("UNIT_POWER_FREQUENT")
	else
		frames.powerBar.bar:Hide()
		self:UnregisterEvent("UNIT_POWER_FREQUENT")
	end
end

function CompactRunes:RUNE_POWER_UPDATE(event, runeIndex, isEnergized)
	self:UpdateRuneCooldown(runeIndex, isEnergized)
end

function CompactRunes:RUNE_TYPE_UPDATE(event, runeIndex)
	self:UpdateRuneType(runeIndex)
end

function CompactRunes:UNIT_POWER_FREQUENT(event, unit, type)
	if unit == "player" then -- and type == "RUNIC_POWER" then
		self:UpdateRunicPower()
	end
end

function CompactRunes:UpdateRunicPower()
	frames.powerBar:SetValue(_G.UnitPower("player", 6))
end

function CompactRunes:UpdateRunicPowerMax()
	frames.powerBar.bar:SetMinMaxValues(0, _G.UnitPowerMax("player", 6))
end

function CompactRunes:UNIT_MAXPOWER(event)
	self:UpdateRunicPowerMax()
end

function CompactRunes:UNIT_MAXHEALTH(event, unit)
	if unit == "target" then
		self:UpdateTargetMaxHealth()
	end
end

function CompactRunes:UNIT_HEALTH(event, unit)
	if unit == "target" then
		self:UpdateTargetHealth()
	end
end

function CompactRunes:UNIT_LEVEL(event, unit)
	if unit == "player" then
		self:CheckTalents()
	end
end

function CompactRunes.CheckDisplayAlpha()
	CompactRunes:SetDisplayAlpha()
end

function CompactRunes:SetDisplayAlpha(combat)
	local alpha = 0
	local inCombat = combat ~= nil and combat or _G.UnitAffectingCombat("player")
	if inCombat then
		alpha = self.db.profile.combatAlpha
	else
		if (addon.depletedRunes + addon.fullyDepletedRunes) > 0 and
			self.db.profile.oocRechargeAlphaEnabled then
			alpha = self.db.profile.oocRechargeAlpha
			-- Start timer to check if runes have recharged and move to OOC Alpha
			_G.C_Timer.After(5, self.CheckDisplayAlpha)
		else
			alpha = self.db.profile.ooc_alpha
		end
	end
	frames.main:SetAlpha(alpha)
	frames.procInd:SetDesiredAlpha()
end

function CompactRunes:PLAYER_ALIVE()
	addon:FireCallback("PlayerAlive")
end

function CompactRunes:PLAYER_DEAD()
	addon:FireCallback("PlayerDead")
end

function CompactRunes:PLAYER_REGEN_DISABLED()
	-- Player has entered combat.
	self:SetDisplayAlpha(true)
	self.LastFightStats:Reset()
	if self.db.profile.rangeInd then
		frames.main:SetScript("OnUpdate", onUpdateInd)
	end
	local bar = self.cdbars[self.currentSpec]
	if bar and bar.SetAlpha then bar:SetAlpha() end
	addon:FireCallback("CombatStart")
end

function CompactRunes:PLAYER_REGEN_ENABLED()
	-- Player has left combat.
	self:SetDisplayAlpha()
	if self.db.profile.rangeInd and not self.targetHostile then
		frames.main:SetScript("OnUpdate", nil)
		frames.main.outOfRange = false
		frames.main:UpdateInd()
	end
	local bar = self.cdbars[self.currentSpec]
	if bar and bar.SetAlpha then bar:SetAlpha() end
	addon:FireCallback("CombatEnd")
end

function CompactRunes:PLAYER_TARGET_CHANGED(unit)
	self.targetHostile = _G.UnitExists("target") and 
		_G.UnitCanAttack("player", "target") and
		not _G.UnitIsDead("target")
	if self.targetHostile and self.db.profile.rangeInd then
		frames.main:SetScript("OnUpdate", onUpdateInd)
	else
		frames.main:SetScript("OnUpdate", nil)
		frames.main.outOfRange = false
		frames.main:UpdateInd()
	end

	self:UpdateTargetMaxHealth()
	self.currentTarget = _G.UnitGUID("target")
	addon.diseaseCount = -1

	self:UNIT_AURA("UNIT_AURA", "target")
	self:CheckPlagueLeech()
	self:CheckSoulReaper()
end

function addon.SetupSpellAuras()
	addon.StandardDiseases = {
		[SpellNames["Frost Fever"]] = frames.ffBar,
		[SpellNames["Blood Plague"]] = frames.bpBar,
	}
	addon.NPDiseases = {
		[SpellNames["Necrotic Plague"]] = frames.npBar,
	}
	addon.petBuffs = {
		SpellNames["Dark Transformation"],
		SpellNames["Shadow Infusion"],
	}
end

function CompactRunes:UNIT_AURA(event, unit)
	local name, rank, icon, count, dispelType, duration, expires, 
		caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
		isBossDebuff, castByPlayer, value1, value2, value3 

	if unit == "target" then
		local diseases = self.necroticPlague and addon.NPDiseases 
			or addon.StandardDiseases
		local diseaseCount = 0

		for spell, bar in pairs(diseases) do
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
			= UnitDebuff("target", spell, nil, "PLAYER")

			if name and bar.db.enabled then
				diseaseCount = diseaseCount + 1
				bar.bar.active = true
				local timeLeft = expires - GetTime()
				if spell == SpellNames["Necrotic Plague"] then
					if bar.db.primary == "Stacks" then
						bar.bar.value:SetText(tostring(count))
					else
						bar.bar.secondary:SetText(tostring(count))
					end
					if bar.db.progress == "Time" then
						bar.bar:SetMinMaxValues(0, duration)
					elseif bar.db.progress == "Stacks" then
						bar:SetValue(count)
					end
				else
					bar.bar:SetMinMaxValues(0, duration)
				end
				bar.bar.timer = timeLeft
				bar.bar:Show()
				if spell == SpellNames["Necrotic Plague"] then
					bar.bar:SetScript("OnUpdate", onUpdateNecrotic)
				else
					bar.bar:SetScript("OnUpdate", onUpdateDisease)
				end
			else
				bar.bar.active = false 
				bar.bar.timer = 0
				bar.bar:SetScript("OnUpdate", nil)
				bar.bar:Hide()
			end
		end

		if diseaseCount ~= addon.diseaseCount then
			addon.diseaseCount = diseaseCount
			self:CheckPlagueLeech()
		end

		name, rank, icon, count, dispelType, duration, expires, 
		caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
		isBossDebuff, castByPlayer, value1, value2, value3 
		= _G.UnitDebuff("target", SpellNames["Soul Reaper"], nil, "PLAYER")
		local old = self.targetSoulReaper
		if name then
			self.targetSoulReaper = true
		else
			self.targetSoulReaper = false
		end
		if old ~= self.targetSoulReaper then
			self:UpdateProcIcon()
		end
	elseif unit == "player" then
		if self.db.profile.bars.BloodShieldBar.enabled then
			local bsBar = frames.bsBar
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
				= _G.UnitBuff("player", SpellNames["Blood Shield"])
			if name then
				bsBar.bar.timer = expires - GetTime()
				bsBar.bar:SetMinMaxValues(0, duration)
				bsBar.bar.value:SetText(FormatNumber(value1 or 0))
				bsBar.bar.active = true 
				bsBar.bar:Show()
				bsBar.bar:SetScript("OnUpdate", onUpdateBloodCharge)
			else
				bsBar.bar.active = false 
				bsBar.bar.timer = 0
				bsBar.bar:SetScript("OnUpdate", nil)
				bsBar.bar:Hide()
			end
		end

		if self.enabled.BoneShieldBar then
			local bar = frames.boneBar
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
				= _G.UnitBuff("player", SpellNames["Bone Shield"])
			local state = nil
			if name then
				state = 0
				local timeLeft = expires - GetTime()
				if bar.db.progress == "Charges" and timeLeft > bar.bar.timer + 10 then
					bar.bar:SetMinMaxValues(0, count)
				end
				bar.bar.timer = timeLeft
				bar.bar.active = true 
				if count ~= bar.bar.count then
					bar.bar.value:SetText(tostring(count))
				end
				bar.bar.count = count
				bar.bar.recharging = false
				bar.bar.recharging = 0
			else
				bar.bar.active = false 
				bar.bar.timer = 0
				bar.bar.count = 0
				local start, duration, enabled = 
					GetSpellCooldown(SpellIds["Bone Shield"])
				if duration > 0 then
					state = 1
					bar.bar.recharging = true
					bar.bar.recharge = duration - (GetTime() - start)
				else
					state = 2
					bar.bar.recharging = false
					bar.bar.recharging = 0
				end
			end

			-- Update the display if state has changed
			if state ~= bar.bar.state then
				if state == 0 then
					-- Active
					if bar.db.progress == "Charges" then
						bar.bar:SetMinMaxValues(0, bar.bar.count)
					elseif bar.db.progress == "Time" then
						bar.bar:SetMinMaxValues(0, duration)
					else
						bar.bar:SetMinMaxValues(0, 1)
						bar.bar:SetValue(1)
					end
					bar.bar:SetAlpha(1)
					bar.bar.rechargeTime:Hide()
					bar.bar.value:Show()
					bar.bar:Show()
					bar.bar:SetScript("OnUpdate", onUpdateBoneShield)
				elseif state == 1 then
					-- Recharging
					if bar.db.showCooldown then
						bar.bar.value:Hide()
						bar.bar.rechargeTime:Show()
						bar.bar:SetAlpha(0.5)
						bar.bar:SetValue(_G.select(2, bar.bar:GetMinMaxValues()))
						bar.bar:SetScript("OnUpdate", onUpdateBoneShield)
						bar.bar:Show()
					else
						bar.bar:Hide()
					end
				else 
					-- Ready
					bar.bar.rechargeTime:Hide()
					bar.bar.value:Show()
					bar.bar.value:SetText("-")
					if bar.db.showReady then
						bar.bar:SetAlpha(0.5)
						bar.bar:SetValue(_G.select(2, bar.bar:GetMinMaxValues()))
						bar.bar:SetScript("OnUpdate", nil)
						bar.bar:Show()
					else
						bar.bar:Hide()
					end
				end
			end
			bar.bar.state = state
		end

		if self.hasBloodTap then
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
				= _G.UnitBuff("player", SpellNames["Blood Charge"])
			if name then
				if expires > self.bcExpires then
					if self.bloodCharges >= addon.MaxBloodCharges - 1 then
						self.TotalStats.cappedBC = self.TotalStats.cappedBC + 1
						self.LastFightStats.cappedBC = self.LastFightStats.cappedBC + 1
					end
				end

				local event = self.db.profile.events["Blood Charge"]
				if event.enabled and event.sound and event.sound ~= "None" and 
					count >= event.count and self.bcExpires ~= expires then
					_G.PlaySoundFile(LSM:Fetch("sound", event.sound), 
						event.soundChannel or addon.defaultSoundChannel)
				end

				self.bloodCharges = count
				self.bcExpires = expires
			else
				self.bloodCharges = 0
				self.bcExpires = 0
			end

			if self.db.profile.bars.BloodChargeBar.enabled then
				local bcBar = frames.bcBar
				if name then
					if count >= 10 then
						bcBar.bar:SetScale(1.15)
					else
						bcBar.bar:SetScale(1.0)
					end
					bcBar.bar.timer = expires - GetTime()
					bcBar.bar:SetMinMaxValues(0, duration)
					bcBar.bar.value:SetText(tostring(count))
					bcBar.bar.active = true 
					bcBar.bar:Show()
					bcBar.bar:SetScript("OnUpdate", onUpdateBloodCharge)
				else
					bcBar.bar.active = false 
					bcBar.bar.timer = 0
					bcBar.bar:SetScript("OnUpdate", nil)
					bcBar.bar:Hide()
				end
			end
		end

		if self.db.profile.procs then
			local procInd = frames.procInd
			local procPriority = self.db.profile.priorityList[self.currentSpec]
			local enabled = self.db.profile.enabled[self.currentSpec]
			if procPriority and enabled then
				for i, spell in ipairs(procPriority) do
					if not self.specialProcs[spell] then
						name, rank, icon, count, dispelType, duration, expires, 
						caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
						isBossDebuff, castByPlayer, value1, value2, value3 
						= _G.UnitBuff("player", SpellNames[spell])
						if name then
							self.currentProcs[spell] = (count or 1)
						else
							self.currentProcs[spell] = nil
						end
						if spell == "Blood Charge" then
							local stacks = name and (count or 0) or 0
							if stacks >= 10 then
								self.currentProcs["Blood Charge (10+)"] = count
								self.currentProcs["Blood Charge (5-9)"] = nil
							elseif stacks >= 5 then
								self.currentProcs["Blood Charge (5-9)"] = count
								self.currentProcs["Blood Charge (10+)"] = nil
							else
								self.currentProcs["Blood Charge (10+)"] = nil
								self.currentProcs["Blood Charge (5-9)"] = nil
							end
						end
					end
				end
				if enabled["Dark Transformation"] then
					name, rank, icon, count, dispelType, duration, expires, 
					caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
					isBossDebuff, castByPlayer, value1, value2, value3 
					= _G.UnitBuff("player", SpellNames["Shadow Infusion"])
					if name and count and count >= 5 then
						self.currentProcs["Dark Transformation"] = 1
					else
						self.currentProcs["Dark Transformation"] = nil
					end
				end
			end
			self:UpdateProcIcon()
		end
	elseif unit == "pet" then
		for i, spell in pairs(addon.petBuffs) do
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
			= _G.UnitBuff("pet", spell, nil, "PLAYER")
			if name then break end
		end

		local dtBar = frames.dtBar
		if name then
			dtBar.bar.dt = (name == SpellNames["Dark Transformation"])
			dtBar.altcolor = dtBar.bar.dt
			dtBar.bar.timer = expires - GetTime()
			local dtdb = self.db.profile.bars.DarkTransformationBar
			local bc, tc
			if dtBar.altcolor then
				bc = dtdb.alt_color
				tc = dtdb.alt_textcolor
			else
				bc = dtdb.color
				tc = dtdb.textcolor
				dtBar.bar.value:SetText(tostring(count))
			end
			dtBar.bar:SetMinMaxValues(0, duration)
			dtBar.bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
			dtBar.bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
			if self.db.profile.bars.DarkTransformationBar.enabled then
				dtBar.bar.active = true 
				dtBar.bar:Show()
				dtBar.bar:SetScript("OnUpdate", onUpdateDarkTrans)
			end
		else
			dtBar.bar.active = false 
			dtBar.bar.timer = 0
			dtBar.bar:SetScript("OnUpdate", nil)
			dtBar.bar:Hide()
		end
	end
end

function CompactRunes:PLAYER_SPECIALIZATION_CHANGED(event)
	self:CheckTalents()
end

function CompactRunes:PLAYER_TALENT_UPDATE(event)
	self:CheckTalents()
end

function CompactRunes:UNIT_SPELLCAST_SUCCEEDED(event, unit, spellName, rank, lineId, spellId)
	-- if unit == "player" then
	-- 	if spellName == SpellNames["Soul Reaper"] then
	-- 		self:ScheduleTimer("CheckSoulReaper", 0.5)
	-- 	end
	-- end
end

function CompactRunes:COMBAT_LOG_EVENT_UNFILTERED(...)
    local event, timestamp, eventtype, hideCaster, 
        srcGUID, srcName, srcFlags, srcRaidFlags, 
        destGUID, destName, destFlags, destRaidFlags, 
        param9, param10, param11, param12, param13, param14, 
        param15, param16, param17, param18, param19, param20

    event, timestamp, eventtype, hideCaster, 
    srcGUID, srcName, srcFlags, srcRaidFlags,
    destGUID, destName, destFlags, destRaidFlags,
    param9, param10, param11, param12, param13, param14, 
    param15, param16, param17, param18, param19, param20 = ...

	if destGUID == addon.playerGUID then
		if eventtype == "SPELL_AURA_APPLIED" then
			if param9 == SpellIds["Runic Corruption"] then
				self.LastFightStats.rcProcs = self.LastFightStats.rcProcs + 1
				self.TotalStats.rcProcs = self.TotalStats.rcProcs + 1
				if frames.main then
					frames.main.rcActive = true
					frames.main:UpdateInd()
				end
			end
		elseif eventtype == "SPELL_AURA_REMOVED" then
			if param9 == SpellIds["Runic Corruption"] then
				if frames.main then
					frames.main.rcActive = false
					frames.main:UpdateInd()
				end
			end
		end
	end
	if srcGUID == addon.playerGUID then
		if eventtype == "SPELL_DAMAGE" or 
			(eventtype == "SPELL_HEAL" and self.regenMethod == "BT") then
			local damage = param12 or 0
			if damage > 0 and self.regenAbilities[param9] ~= nil then
				local regenMethod = self.regenMethod
				self.TotalStats.regenUses[regenMethod] =
					(self.TotalStats.regenUses[regenMethod] or 0) + 1
				self.LastFightStats.regenUses[regenMethod] = 
					(self.LastFightStats.regenUses[regenMethod] or 0) + 1
			end
		--elseif eventtype == "SPELL_AURA_APPLIED" or 
		--	eventtype == "SPELL_AURA_REFRESH" then
		--	if param9 == SpellIds["Frost Fever"] or 
		--		param9 == SpellIds["Blood Plague"] then
		--		if destGUID == self.currentTarget then
		--			self:CheckPlagueLeech()
		--		end
		--	end
		--elseif eventtype == "SPELL_AURA_REMOVED" then
		--	if param9 == SpellIds["Frost Fever"] or 
		--		param9 == SpellIds["Blood Plague"] then
		--		if destGUID == self.currentTarget then
		--			self:CheckPlagueLeech()
		--		end
		--	elseif param9 == SpellIds["Necrotic Plague"] and 
		--		destGUID == self.currentTarget then
		--		self:CheckPlagueLeech()
		--	end
		end
		local srid = self.soulReaperIds[self.currentSpec]
		if eventtype == "SPELL_DAMAGE" and param9 == srid then
			local start, duration, enabled = GetSpellCooldown(srid)

			self:CheckSoulReaper(false, true)
		end
	end
end

function CompactRunes:SPELL_UPDATE_COOLDOWN(event)
	local changed = false
	local ignoreRuneCDs = self.db.profile.ignoreRuneCDs["Procs"]

	--local enabled = self.db.profile.enabled[self.currentSpec]
	for k, v in pairs(self.AbilityProcs) do
		if k and v then
			local start, duration, enabled = 
				GetSpellCooldown(SpellIds[k])
			local ready = (duration < 2)
			local ignore = false
			if not ready and ignoreRuneCDs then
				local grace = self.IgnoreRuneCDs[k]
				if grace and grace > 0 and duration <= grace then
					ignore = true
				end
			end
			if not ignore then
				self.currentProcs[k] = (ready and 1 or nil)
				if self.ready[k] ~= ready then
					changed = true
				end
				self.ready[k] = ready
			end
		end
	end

	if self:CheckSoulReaper(true) then changed = true end

	if changed then
		self:UpdateProcIcon()
	end
end

function CompactRunes:CheckForGhoul()
	if self.currentSpec == "Unholy" and not _G.UnitExists("pet") and 
		not _G.IsMounted() then
		-- Ghoul is missing.
		if not self.currentProcs["Raise Dead"] then
			self.currentProcs["Raise Dead"] = 1
			self:UpdateProcIcon()
		end
	else
		if self.currentProcs["Raise Dead"] then
			self.currentProcs["Raise Dead"] = nil
			self:UpdateProcIcon()
		end
	end
end

function CompactRunes:UNIT_PET(event, unit)
	if unit and unit == "player" then
		self:CheckForGhoul()
	end
end

function CompactRunes:PLAYER_ENTERING_WORLD(event)
	for i = 1, 6 do
		self:UpdateRuneType(i)
		self:UpdateRuneCooldown(i)
	end
	self:UpdateRunicPower()
	addon.playerGUID = _G.UnitGUID("player") or ""
	self:UNIT_AURA("PLAYER_ENTERING_WORLD", "player")
end

local TierSlotIds = {
	["Head"] = _G.GetInventorySlotInfo("HeadSlot"),
	["Shoulder"] = _G.GetInventorySlotInfo("ShoulderSlot"),
	["Chest"] = _G.GetInventorySlotInfo("ChestSlot"),
	["Legs"] = _G.GetInventorySlotInfo("LegsSlot"),
	["Hands"] = _G.GetInventorySlotInfo("HandsSlot"),
}

local TierIds = {
	["T15 DPS"] = {
		["Head"] = {
			[95827] = true,
			[95227] = true,
			[96571] = true,
			},
		["Shoulder"] = {
			[95829] = true,
			[95229] = true,
			[96573] = true,
			},
		["Chest"] = {
			[95825] = true,
			[95225] = true,
			[96569] = true,
			},
		["Legs"] = {
			[95828] = true,
			[95228] = true,
			[96572] = true,
			},
		["Hands"] = {
			[95826] = true,
			[95226] = true,
			[96570] = true,
		},
	},
	["T16 DPS"] = {
		["Head"] = {
			[99057] = true,
			[99194] = true,
			[99337] = true,
			},
		["Shoulder"] = {
			[99059] = true,
			[99187] = true,
			[99339] = true,
			},
		["Chest"] = {
			[99066] = true,
			[99192] = true,
			[99355] = true,
			},
		["Legs"] = {
			[99058] = true,
			[99186] = true,
			[99338] = true,
			},
		["Hands"] = {
			[99067] = true,
			[99193] = true,
			[99336] = true,
		},
	},
	["T18"] = {
		["Head"] = {
			[124332] = true,
			},
		["Shoulder"] = {
			[124344] = true,
			},
		["Chest"] = {
			[124317] = true,
			},
		["Legs"] = {
			[124338] = true,
			},
		["Hands"] = {
			[124327] = true,
		},
	},
}

local TierSlots = {}
for k, v in pairs(TierSlotIds) do
	TierSlots[v] = true
end

local TrackedItemSlotIds = {
	["Trinket1"] = _G.GetInventorySlotInfo("Trinket0Slot"),
	["Trinket2"] = _G.GetInventorySlotInfo("Trinket1Slot"),
}
local TrackedItemSlots = {}
for k, v in pairs(TrackedItemSlotIds) do
	TrackedItemSlots[v] = true
end

local GearChangeTimer = nil
function CompactRunes:CheckGear()
	GearChangeTimer = nil
	local count = 0
	local changed = false

	for tier, ids in pairs(TierIds) do
		count = 0
		for slot, slotid in pairs(TierSlotIds) do
			local id = _G.GetInventoryItemID("player", slotid)
			if ids[slot][id] then
				count = count + 1
			end
		end

		if count ~= addon.tierCount[tier] then
			addon.tierCount[tier] = count
			if self.db.profile.debug and not _G.InCombatLockdown() then
				local fmt = "%s Detected: %d/5"
				self:Print(fmt:format(tier, addon.tierCount[tier]))
			end
			changed = true
		end
	end

	if changed then self:UpdateTierBonus() end
	for name, bar in pairs(self.cdbars) do
		if bar then bar:Update() end
	end

	addon:FireCallback("GearUpdate")
end

function CompactRunes:UpdateTierBonus()
	self:SetSoulReaperPerc()
end

function CompactRunes:PLAYER_EQUIPMENT_CHANGED(event, slot, hasItem)
	if (TierSlots[slot] or TrackedItemSlots[slot]) and not GearChangeTimer then
		GearChangeTimer = self:ScheduleTimer("CheckGear", 1.5)
	end
end
